#!/usr/bin/perl
#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#.IDENTIFICATION  wdb
#.LANGUAGE	  Perl 5, DBI/DBD script
#
#.PURPOSE	  WWW Database Interface.
#
#.AUTHOR	  Bo Frese Rasmussen [ST-ECF]   <bfrasmus@eso.org>
#.MODIFICATIONS   Jeff Rowe                     <beowulf@lava.net>
#		  http://www.lava.net/beowulf/programming/wdb/index.html
#
#.COPYRIGHT (see end of file)
#------------------------------------------------------------------------------

package WDB;
use strict;  # Protects me from a lot of sloppiness.

#-------------------------------------------------------
# Prepare STDOUT and STDERR
#-------------------------------------------------------
open(STDERR,">&STDOUT") || die "Can't dup stdout: $!\n";
select(STDERR); $| = 1; # Make unbuffered.
select(STDOUT); $| = 1; # Make unbuffered.

# Ugly global variables that must be declared.
use vars qw(@fields %field %form @keys $selectlist $wherelist @select_titles);
use vars qw($count_wherelist $sel_th $main_th %in @in $in $debug $sql_count);
use vars qw($junk $user_menu $db_query $tmp_file $output_mode $rowcount);
use vars qw($ACTION $WDB $DBDIR $FORMNAME $KEY_ARGS $ALLOW_MASTER $fromlist);
use vars qw($USE_MOD_PERL $VERSION $SHORT_VERSION $REL_DATE $CONFIG_DIR);
use vars qw($db_insert $db_update $db_delete $insert_th $update_th $delete_th);
use vars qw($result $NO_QUERY_BUTTON %old_values );
use vars qw($GLOBAL_ROWCOUNT $GLOBAL_MAXROWCOUNT);
use vars qw($GLOBAL_USER $GLOBAL_PSWD $GLOBAL_SERVER);

# Variables in wdb.conf that must be declared.
use vars qw($FORMDIR $MAIN_MENU $WDB_HTML $IMAGEDIR $WDB_HTML $QUERY_HELP);
use vars qw($USER $PSWD $SERVER $DBI_FILE $MAXROWCOUNT $ROWCOUNT $DATEFORMAT);
use vars qw($NULL_VALUE @IP_IGNORE %SPEC_NULL $SECURE_WDB $HELPTEXT);
use vars qw($HOME_BUTTON $QUERY_BUTTON $HELP_BUTTON);

#-----------------------------------------------
# Under mod_perl these variables are initialized ONCE when the Apache child is
#  created, and never again.  This is different than those in the init_vars()
#  subroutine which are initialized every time the program is called.
#-----------------------------------------------
BEGIN {
   # Everyone needs to set this value to match their own system configuration.
   $CONFIG_DIR = "/etc/wdb";
# line 50 WDB

   # Auto-detect if we are running under mod_perl or CGI.
   $USE_MOD_PERL = 0;
   if ($ENV{'GATEWAY_INTERFACE'} =~ /CGI-Perl/
       || (exists $ENV{'MOD_PERL'})) {
      $USE_MOD_PERL = 1;
   }

   # Version information.  Don't change this or you'll confuse people. :)
   $VERSION       = "wdb 1.5dm-beta5.4";
   $SHORT_VERSION = "1.5dm-b5.4";
   $REL_DATE      = "13-Dec-1997";
}

use lib "$CONFIG_DIR";
require "wdb.conf";

#--------------------------
# Load database interface.
#--------------------------
require "$DBI_FILE";

# Perl version 5.004 is required for mod_perl.  If you don't use mod_perl, WDB
#  works just fine with 5.003 or better.
require 5.004 if exists $ENV{'MOD_PERL'};

# Use CgiError(), ReadParse(), PrintVariables() from cgi-lib.pl
require	"cgi-lib.pl";

#-----------------------------------------------
# Initialize variables.  Not required under CGI, but a must under mod_perl.
#  Probably a darn good idea at any time.
#-----------------------------------------------
&init_vars;

print "Content-type: text/html\n\n" unless $USE_MOD_PERL; 

#------------------------------
# Parse info from Web Client
#------------------------------
if (defined $ENV{'PATH_INFO'}) {
   ( $junk , $DBDIR, $FORMNAME, $ACTION, $KEY_ARGS ) 
     = split( '/', $ENV{'PATH_INFO'}, 5 );
}

# Use a protected path for $SECURE_WDB to protect updates and deletes.
if ( $WDB eq $SECURE_WDB ) {
   $ALLOW_MASTER = 1;
}
# line 100 WDB

#---------------------
# Main() ...
#---------------------

# The 'dummy' action allows a selectable list of tables and actions.
# Other functionality can be added.
if ( $ACTION eq "dummy" ) {
   &ReadParse(\%in, \@in, \$in);
   $FORMNAME = $in{'form'};
   $ACTION = $in{'action'};
}

# Parse the FDF.
&ParseFile("$FORMDIR/$DBDIR/$FORMNAME.fdf", \@fields, \%field, \@keys, \%form);

# Let the FDF creator set these values to override the global values in
#  wdb.conf if needed for a specific FDF.
$GLOBAL_ROWCOUNT = (defined $form{'ROWCOUNT'} ? $form{'ROWCOUNT'} : $ROWCOUNT);
$GLOBAL_MAXROWCOUNT = 
   (defined $form{'MAXROWCOUNT'} ? $form{'MAXROWCOUNT'} : $MAXROWCOUNT);
$GLOBAL_USER = (defined $form{'USER'} ? $form{'USER'} : $USER);
$GLOBAL_PSWD = (defined $form{'PASSWD'} ? $form{'PASSWD'} : $PSWD);
$GLOBAL_SERVER = (defined $form{'SERVER'} ? $form{'SERVER'} : $SERVER);

# Choose which action to perform.  These actions put up a query form targeted 
#  at the specific function, or perform the requested database action.
if ( $ACTION eq "query_form" || 
     $ACTION eq "form" ) {
    # The $rowcount variable needs to be set every time under mod_perl.
    $rowcount = $GLOBAL_ROWCOUNT;
    &PrintQueryForm;
# These actions perform the database query and return the results.
} elsif ( $ACTION eq "query" || 
	  $ACTION eq "default" ||
	  $ACTION eq "detail" ||
          $ACTION eq "page" ) {
    if ($ACTION eq "detail") { $NO_QUERY_BUTTON = 1; $ACTION = "query"; }
    &DoQuery;
} elsif ($ACTION =~ /insert|update|delete/ && !$ALLOW_MASTER) {
    &CgiError("The current action is not allowed from the specified path.\n");
    goto BAIL;
} elsif ($ACTION eq "update_query_form") {
    if (defined $form{'ALLOW_UPDATE'}) {
       &PrintQueryForm;
    } else {
        print $HELPTEXT;
        &CgiError("This FDF file does not allow updates.\n");
        goto BAIL;
# line 150 WDB
    }
} elsif ($ACTION eq "delete_query_form") {
    if (defined $form{'ALLOW_DELETE'}) {
       &PrintQueryForm;
    } else {
        print $HELPTEXT;
        &CgiError("This FDF file does not allow deletes.\n");
        goto BAIL;
    }
} elsif ($ACTION eq "update_query") {
    if (defined $form{'ALLOW_UPDATE'}) {
       &DoQuery;
    } else {
        print $HELPTEXT;
        &CgiError("This FDF file does not allow updates.\n");
        goto BAIL;
    }
} elsif ($ACTION eq "delete_query") {
    if (defined $form{'ALLOW_DELETE'}) {
       &DoQuery;
    } else {
        print $HELPTEXT;
        &CgiError("This FDF file does not allow deletes.\n");
        goto BAIL;
    }
# Display an insert form to add new records.
} elsif ( $ACTION eq "insert_form" ) {
    if ($form{'ALLOW_INSERT'}) {
        &PrintInsertForm;
    } else {
	print $HELPTEXT;
	&CgiError("This FDF file does not allow inserts.\n"); 
	goto BAIL;
    }
# Display a form with the information from a previous select to allow updates.
} elsif ( $ACTION eq "update_form" ) {
    if (defined $form{'ALLOW_UPDATE'}) {
       &PrintProcessingForm;
    } else {
	print $HELPTEXT;
	&CgiError("This FDF file does not allow updates.\n"); 
	goto BAIL;
    }
# Display a form with the information from a previous select to allow a delete.
} elsif ( $ACTION eq "delete_form" ) {
    if (defined $form{'ALLOW_DELETE'}) {
       &PrintProcessingForm;
    } else {
	print $HELPTEXT;
# line 200 WDB
	&CgiError("This FDF file does not allow deletes.\n"); 
	goto BAIL;
    }
# Perform the actual insertion of data.
} elsif ( $ACTION eq "insert" ) {
    if ($form{'ALLOW_INSERT'}) {
        &DoInsert;
    } else {
	print $HELPTEXT;
	&CgiError("This FDF file does not allow inserts.\n"); 
	goto BAIL;
    }
# Perform the actual update of data.
} elsif ( $ACTION eq "update" ) {
    if (defined $form{'ALLOW_UPDATE'}) {
       &DoUpdate;
    } else {
	print $HELPTEXT;
	&CgiError("This FDF file does not allow updates.\n"); 
	goto BAIL;
    }
# Peform the actual deletion of data.
} elsif ( $ACTION eq "delete" ) {
    if (defined $form{'ALLOW_DELETE'}) {
       &DoDelete;
    } else {
	print $HELPTEXT;
	&CgiError("This FDF file does not allow deletes.\n"); 
	goto BAIL;
    }
# The requested action is unknown.  Bail out.
} else {
    print $HELPTEXT;
    &CgiError("Unknown action in PATH_INFO: $junk, $DBDIR, $FORMNAME, $ACTION");
    goto BAIL;
}

#------------------------------------------------------------------------------
# There is no good way to exit this program immediately when run under mod_perl
#  so this is the best I could come up with.  Using exit() or die() produces
#  a server error, when all I want to do is produce an error message and bail
#  out gracefully.  I also want to make sure the database connections get
#  closed if an error occurs and I have no idea which kind of connection has
#  been made, hence the kludge below.
#------------------------------------------------------------------------------
BAIL: &dbi_disconnect($db_query, \$main_th);
      &dbi_disconnect($db_insert, \$insert_th);
      &dbi_disconnect($db_update, \$update_th);
      &dbi_disconnect($db_delete, \$delete_th);
# line 250 WDB
      return if $USE_MOD_PERL;

#------------------------------------------------------------------------------
# These are variables that MUST be initialized under mod_perl.  They are
#  reset EVERY time the script is run.
# NOTE: if your database requires environment variables to be set before you
#  can run database software, do that here too.
#------------------------------------------------------------------------------
sub init_vars {
   $user_menu = $sql_count = $KEY_ARGS = $db_query = $output_mode = "";
   $selectlist = $wherelist = $count_wherelist = $sel_th = $main_th = "";
   $ACTION = $FORMNAME = $DBDIR = $in = "";
   %field = %form = %in = %old_values = ();
   @fields = @keys = @select_titles = @in = ();
   $rowcount = 50;
   $NO_QUERY_BUTTON = 0;
   $GLOBAL_ROWCOUNT = $GLOBAL_MAXROWCOUNT = 0;
   $GLOBAL_USER = $GLOBAL_PSWD = $GLOBAL_SERVER = "";

   # Database stuff for Informix.  Your milage may vary.
#   $ENV{'SQLEXEC'}       = "/usr/local/informix/lib/sqlturbo";
#   $ENV{'INFORMIXDIR'}   = "/usr/local/informix";
#   $ENV{'DBDATE'}        = "MDY2";
}

#
#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#.PURPOSE	Parse a form definition file ('FDF' file)
#
#.REMARKS       Expects 5 parameters: the FDF name, an empty array to contain
#                the field names, an empty hash to contain field parameters,
#                an empty array to contain key field information, and an empty
#                hash to contain form parameters.
#               The last 4 parameters need to be passed in by reference.
#
#.RETURNS	True, or not at all ....
#------------------------------------------------------------------------------
sub ParseFile 
{
    my( $file, $fields, $field, $keys, $form ) = @_;
    my( $current, $key, $val, $lineno, $void, $old_attr_key, $old_attr_type );

    # Open the FDF.
    if (!open( FDF, $file )) {
       &CgiError("Can't open $file: $!\n");
       goto BAIL;
    }

    $lineno = 0;
# line 300 WDB
    while ( <FDF> ) {	  
	$lineno++;

	#-----------------------------------------------------------------
	# Parse a line of the form : "key=value". Comments and unnessesary
	# blanks are skipped. (blanks in the middle of values are kept!)
	# The key should start at the 1. character on the line - lines 
	# starting with whitespace are interpreted as continuation lines.
	#-----------------------------------------------------------------
	next if /^#.*$/;
	($key, $void, $val ) = 
		/^(\w+)\s*(=\s*((\s*\".*\"|\s*[^\s\n]+)+))?\s*(#.*)?$/;
		#/^(\w+)\s*(=\s*((\s*\".*\"|\s*[^#\s\n]+)+))?\s*(#.*)?$/;
		# /^(\w+)\s*(=\s*((\s*[^#\"\s\n]+|\s*\".*\")+))?\s*(#.*)?$/;

	if ( ! $key ) {   # Continuation line ...
	    # The original would screw up '#' signs in URLs - JPR
	    ($val) = /^\s*((\s*[^\s\n]+)+)\s*(#.*)?$/;
	    #($val) = /^\s*((\s*[^#\s\n]+)+)\s*(#.*)?$/;
	    next unless $val;
	    if ( $old_attr_type eq 'form' ) {
		$form->{$old_attr_key} .= "\n$val";
	    } else {
		$$field{$old_attr_key} .= "\n$val";
	    }
	    next;
	}

	#--------------------------------------------------
	# Beginning of a new FIELD definition ?
	#--------------------------------------------------
	if ( $key eq "FIELD" ) {
	    if ( $current ) {	# If not first field, check old field
		if ( ! $$field{$current,'label'} ) {
		    $$field{$current,'label'} = $current;
		}
	    }
	    $current = $val;
	    push( @$fields, $val );
	    next;
	}

	#--------------------------
	# Is it a FORM attribute ?
	#--------------------------
	if ( ! $current ) {	  # Defs' before first field are form defs'
           #### CHANGED for Perl5 ###
           $val = 'TRUE' unless defined $val;
           ####
# line 350 WDB
	   $form->{$key} = $val;	
	   $old_attr_key = $key;
	   $old_attr_type = 'form';
	   next;
	}

	#--------------------------
	# It's a FIELD attribute.
	#--------------------------
	if ( $key eq "key" ) {
	    unshift( @$keys, $current);
	} 
        #### CHANGED for Perl5 ###
        $val = 'TRUE' unless defined $val;
        ####
	$$field{$current,$key} = $val;
	$old_attr_key   = "$current$;$key";
	$old_attr_type  = 'field';
    }

    #--------------------------------------------------------
    # If a 'PERL' attribute exists in the form : Evaluate it
    #--------------------------------------------------------
    if ( $form->{'PERL'} ) {
	eval $form->{'PERL'};
	if ( $@ ) {
	    &CgiError("Error in the form's PERL attribute :<p>\n $@\n");
	    goto BAIL;
	}
    }
    return 1;
}
#
#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#.PURPOSE	Print an HTML query form for selecting records from a database.
#
#.REMARKS       Expects no parameters.
#
#.RETURNS	True, or not at all ....
#------------------------------------------------------------------------------
sub PrintQueryForm 
{
    my ( $max, $notfirstfield, $f, $value, $title, $rows, $default, $db_sel );
    my ( $tablename , $fieldname, $sql_query, @select_fields, $list_format );
    my ( %myrow, $ef, $length, $search_key, $list_size, @extrafields );
    my ( $fieldnames, $num_rows, $help, $opname, $return_code, $result );
    my ($enum_values);

    #-----------------------------
# line 400 WDB
    # Print HTML header stuff
    #-----------------------------
    &write_header;

    if (@fields < 1) { goto BAIL; }

    &write_menu;

    if (!defined $form{'NO_MAX_RECORDS_BOX'}) {
       print "Return a maximum of ".
	     "<INPUT VALUE=$rowcount SIZE=3 NAME=max_rows_returned> records";
       print " per page" if defined $form{'ROW_PAGING'};
       print " from this query.<BR>";
    } else {
       print "<INPUT TYPE=hidden VALUE=$rowcount NAME=max_rows_returned>";
    }

    # Let the user see how many records match a query without doing the query.
    if ($form{'MATCH_OPTION'}) {
       print "<HR>";
       print "<INPUT TYPE=\"radio\" NAME=\"purpose\" VALUE=\"full\" CHECKED> ".
             "Report Full Detail";
       print "<BR>";
       print "<INPUT TYPE=\"radio\" NAME=\"purpose\" VALUE=\"count\"> Report ".
             "Number of Matches Only";
       print "<BR>";
    }
    if (defined $form{'RECORD_COUNT_FILE'}) {
       print "<HR>";
       &count_records;
    }
    print "<HR>";
    print "\n<pre>";

    #------------------------
    # Print fields
    #------------------------
    $max = 0;
    foreach $f ( @fields ) {
       next if defined $field{$f,'no_query'} 
	       && ($ACTION eq "query_form" || $ACTION eq "form");
       next if defined $field{$f,'hidden'};

       $max= length($field{$f,'label'}) if length($field{$f,'label'}) > $max;
    }

    foreach $f ( @fields ) {
	next if defined $field{$f,'hidden'};

# line 450 WDB
	# Generate a list of extra fields if they are needed.
	if ( defined $field{$f,'no_query'} ) {
	    if ( ! defined $field{$f,'no_tab'}
	         && ! defined $field{$f,'forcetab'} ) { 
	        push( @extrafields, $f );
	    }
	    next;
	}
	if (defined $field{$f,'form_buttons'}) {
	   &write_menu;
	} else {
	   print "\n" if ( $notfirstfield++ && !defined $field{$f,'sameline'} );
	}
	print &eval_attr($f,'html') if defined $field{$f,'html'};

	#---------------------------------
	# Checkbox : In tabular output?  The checkboxes should not appear on
	#  update or delete query forms because all fields are returned from
	#  those queries for data security reasons.
	#---------------------------------
	if (defined $form{'NO_CHECKBOXES'}) {
	   print "<INPUT type=hidden value=1 name=\"tab_$f\">" 
	      if ! defined $field{$f,'no_tab'}; 
	} elsif ( ($ACTION eq "query_form" || $ACTION eq "form")
	     && !defined $form{'FORCE_FULL'} ) {
	   print "<INPUT type=checkbox ";
	   print "checked" if ! defined $field{$f,'no_tab'};
	   print " name=\"tab_$f\">";
	}

	#---------------------------------
	# Label
	#---------------------------------
	print "<b>";
	if ( $field{$f,'help'} ) {
	    print "help field : $field{$f,'help'}<BR>\n" if $debug;
	    $help = &eval_attr($f,'help');
	    print "<a href=\"";
	    print $form{'DOCURL'} if ( $help =~ /^#.*/ );
	    print $help . "\">";
	    print $field{$f,'label'};
	    print "</a>";
	} else {
	    print $field{$f,'label'};
	}

	if ( ! defined $field{$f,'sameline'} ) {
	    print "." x ($max - length($field{$f,'label'})) ;
	    print ":";
	}
	print "</b>";

        #------------------
# line 500 WDB
        # Logical Operators.  Lets the user pick things like 'equals',
	#  'contains', '>=', '<', etc. from a pulldown list instead of
	#  expecting them to read the query help instructions.
        #------------------
        if ( $field{$f,'operators'} ) {
            $opname = $f."_op";
            print "<SELECT name=\"$opname\">";
            foreach ( split(',',$field{$f,'operators'}) ) {
                ( $value, $title ) = /^\s*(.*)!(.*)$/;
                print "<OPTION value=\"$value\"> $title\n";
            }
            print "</SELECT>";
        }

	#--------------------------------------------
	# Query field : Determine the display format.
	#--------------------------------------------
	if ( $field{$f,'enum'} ) {
	    print "<SELECT name=\"$f\"";
	    print " MULTIPLE" if ( defined $field{$f,'multiple'});
	    print " SIZE=$field{$f,'size'}" 
	       if (defined $field{$f,'size'});
	    print ">";
	    if (defined $field{$f,'enum_perl'}) {
	       $enum_values = eval $field{$f,'enum_perl'};
               if ( $@ ) {
                   &CgiError("Error in $f"."'s 'enum_perl' attribute:<p> $@\n");
	           goto BAIL;
               }
	    } else {
	       $enum_values = $field{$f,'enum'};
	    }
	    foreach ( split(',',$enum_values) ) {
		( $value, $title ) = /^\s*(.*)=(.*)$/;
		next if $value eq '%' && defined $field{$f,'multiple'};
		print "<OPTION value=\"$value\"> $title\n";
	    }
	    print "</SELECT>";
        } elsif ( defined $field{$f, 'makelist'} ) {
            @select_fields = %myrow = ();
            $result = &dbi_connect( $GLOBAL_USER, $GLOBAL_PSWD, $GLOBAL_SERVER, 
	                            $form{'DATABASE'}, \$db_sel );
            if ($result) {
               print $result;
               &dbi_disconnect($db_sel, \$sel_th);
               goto BAIL;
            }

            ( $tablename , $fieldnames, $search_key )
# line 550 WDB
               = split('%',$field{$f,'makelist'});
	    if (!defined $tablename || !defined $fieldnames) {
	       print "Incorrectly formed 'makelist' attribute.<BR>\n";
               &dbi_disconnect($db_sel, \$sel_th);
	       goto BAIL;
	    }
	    if (defined $field{$f,'list_format'}) {
	       $list_format = $field{$f,'list_format'};
	    } else {
	       $list_format = "picklist";
	    }
            ( $value, $title ) = split(',', $fieldnames);
            push( @select_fields, $value );
            if (!defined $title || $title eq "") {
               $title = $value;
            } else {
               push( @select_fields, $title );
            }

            $sql_count = "select count(*) from $tablename";
            $sql_count .= " where $value = $search_key" if defined $search_key;
            if ( ($num_rows = 
	          &dbi_countrows($db_sel, $sql_count, @select_fields)) > 0 ) {

               $sql_query = "select distinct $fieldnames from $tablename";
               $sql_query .= " where $value = $search_key" 
	          if defined $search_key;
               $sql_query .= " order by $title";
               &dbi_dosql($db_sel, \$sel_th, $sql_query);

               if ( %myrow = &dbi_nextrow($sel_th, @select_fields) ) {
		   print "<BR>\n" if defined $field{$f, 'list_linebreak'};
                   print "<UL>\n" if defined $field{$f, 'list_indent'};
                   if (($list_format eq "radio" || $list_format eq "checkbox")
		        && defined $field{$f, 'list_first_element'}) {
		      print "\n" if !defined $field{$f, 'list_oneline'};
                      print "<INPUT TYPE=\"$field{$f,'list_format'}\"".
		            " name=\"$f\" CHECKED".
		            " VALUE=\"$field{$f, 'list_first_element'}\">".
                            " $field{$f, 'list_first_element'}  ";
                   } elsif ($list_format eq "picklist") {
		      if (defined $field{$f, 'list_size'}) {
			 $list_size = $field{$f, 'list_size'};
		      } else {
			 $list_size = 4;
		      }
                      print "<SELECT name=\"$f\"";
		      print " MULTIPLE" if defined $field{$f, 'list_multi'};
                      print " SIZE=$list_size"
# line 600 WDB
                         if defined $field{$f, 'list_size'};
                      print ">\n";
                      print "<OPTION value=\"%\"";
		      print " SELECTED" if !defined $field{$f, 'list_multi'};
		      print "> ";
		      print "$field{$f, 'list_first_element'}"
		         if defined $field{$f, 'list_first_element'};
		      print "\n";
                   }
                   do {
                      if ($list_format eq "radio" 
		          || $list_format eq "checkbox") {
                         if ( defined $field{$f, 'list_oneline'} ) {
		            print "\n";
                         } elsif ( defined $field{$f, 'list_indent'} ) {
                            print "</UL><UL>";
                         }
                         print "<INPUT TYPE=\"$field{$f,'list_format'}\" ".
			       "name=\"$f\" ".
                               "VALUE=\"$myrow{$value}\"> $myrow{$title}  ";
                      } else {
                          print "<OPTION value=\"";
			  print "$field{$f, 'list_match'}"
			     if defined $field{$f, 'list_match'};
			  print "$myrow{$value}\">".
                                "$myrow{$title}\n";
                      }
                   } while ( %myrow = &dbi_nextrow($sel_th, @select_fields) );

                   if ($list_format eq "picklist") {
                      print "</SELECT>";
                   }
                   print "</UL>" if defined $field{$f, 'list_indent'};
               }
            }
            @select_fields = ();
            &dbi_disconnect($db_sel, \$sel_th);
	#--------
	# Default text box.
	#--------
        } elsif ( defined $field{$f, 'textarea'} ) {
	   print "<TEXTAREA NAME=\"$f\"";
	   print " $field{$f, 'textarea'}" if $field{$f, 'textarea'} ne "TRUE";
	   print ">";
	   if ( $field{$f,'default'} ) {
	      $default = &eval_attr($f,'default');
	   } else {
	      $default = "";
	   }
# line 650 WDB
           $default =~ s/^\s*//;
           $default =~ s/\s*$//;
	   print $default;
	   print "</TEXTAREA>";
	} else {
	    if ( $field{$f,'length'} > (55-$max) ) {
		$length = (55-$max) ;
	    } else {
		$length = $field{$f,'length'};
	    }
	    if ( $field{$f,'default'} ) {
		$default = &eval_attr($f,'default');
	    } else {
		$default = "";
	    }
	    print "<INPUT name=\"$f\" size=$length ";
	    print "MAXLENGTH=$field{$f,'max_length'} " 
	       if defined $field{$f,'max_length'};
	    print "value=\"$default\" >";
	}
	#-------------
	# Unit label
	#-------------
	print " $field{$f,'unitlabel'}" if ( $field{$f,'unitlabel'} );
    }

    #------------------------------
    # Print the rest of the page.
    #------------------------------
    print "</pre> ";
    if ( @extrafields ) {
	print "<hr><a href=\"$WDB_HTML/wdb_query_help.html#ctrl\">";
	print "Extra columns on tabular output</a> : ";
	print "<SELECT name=\"extra_columns\" size=8 multiple>\n";
	foreach $ef ( @extrafields ) {
	    print "<option value=$ef> $field{$ef,'label'}\n";
	}
	print "</SELECT>\n";
    }

    #----------------
    # Can the user select the output mode?  If the FDF defines FORCE_FULL, the
    #  output can only be displayed in full-screen mode.  If the action is
    #  an update or delete query, the output can only be displayed in tabular
    #  output.  Otherwise, let the user choose.
    #----------------
    if ( $ACTION eq "form" || $ACTION eq "query_form" ) {
	if ( defined $form{'FORCE_FULL'} ) {
           print "<INPUT type=hidden name=\"full_screen_mode\" value=\"yes\">";
# line 700 WDB
	} else {
	   print "<HR>";
           print "<INPUT type=checkbox name=\"full_screen_mode\">";
           print "Use full-screen output even if more than one row is returned".
		 ".<br>";
	}
    }

    &write_menu;

    print "</FORM>\n";

    #---------------
    # Let users keep a visitor count file for their Web pages.
    #---------------
    if (defined $form{'VISITOR_COUNT_FILE'}) {
       $return_code = &count_visitor;
       if ($return_code == -1) { print "Visitor count failed.\n"; }
    }

    #---------------
    # Some people don't want the WDB tagline on the query form.
    #---------------
    &html_tail unless defined $form{'NO_TAIL'};
    print "</BODY></HTML>\n";
}
#
#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#.PURPOSE	Print an HTML insert form for inserting data into the database.
#
#.REMARKS       Expects no parameters
#
#.RETURNS	True, or not at all ....
#------------------------------------------------------------------------------
sub PrintInsertForm 
{
    my( $sec, $min, $hour, $mday, $mon, $year, $wday, $yday, $isdst, $help );
    my( $max, $notfirstfield, $f, $value, $title, $rows, $default );
    my( $time_stamp, $length, $err );
    my ($enum_values);

    #-----------------------------
    # Print HTML header stuff
    #-----------------------------
    &write_header;

    if (@fields < 1) { goto BAIL; }

    &write_menu;
# line 750 WDB
    print "\n<pre>";

    #------------------------
    # Print fields
    #------------------------
    $max = 0;

    #---------------------------------
    # This finds the longest field for formatting purposes.
    #---------------------------------
    foreach $f ( @fields ) {
       next if defined $field{$f,'serial_field'};

       if ( (! defined $field{$f,'no_insert'} ) 
	   && (! defined $field{$f,'hidden'} ) ) {
	  $max= length($field{$f,'label'})+1 
	    if length($field{$f,'label'})+1 > $max;
       }
    }

    foreach $f ( @fields ) {
        next if defined $field{$f,'serial_field'};
	next if defined $field{$f, 'hidden'};
	if (defined $field{$f, 'no_insert'}) {
           if ( $field{$f,'default'} ) {
              $default = eval $field{$f,'default'};
              if ( $@ ) {
                  &CgiError("Error in $f"."'s 'default' attribute:<p> $@\n");
	          goto BAIL;
              }
	   } else {
	      $default = "";
	   }
	   print "<INPUT name=\"$f\" type=hidden value=\"$default\">";
	   next;
        }

	if (defined $field{$f,'form_buttons'}) {
	   &write_menu;
	} else {
	   print "\n" if ( $notfirstfield++ && !defined $field{$f,'sameline'} );
	}
	print &eval_attr( $f,'html') if defined $field{$f,'html'};

	#---------------------------------
	# Label
	#---------------------------------
	print "<b>";
	if ( $field{$f,'help'} ) {
	    print "help field : $field{$f,'help'}<BR>\n" if $debug;
            $help = eval $field{$f,'help'};
            if ( $@ ) {
               &CgiError("Error in $f"."'s  'help' attribute :<p>\n $@\n");
# line 800 WDB
               goto BAIL;
            }
	    print "<a href=\"";
	    print $form{'DOCURL'} if ( $help =~ /^#.*/ );
	    print $help . "\">";
	    print $field{$f,'label'};
	    print "</a>";
	} else {
	    print $field{$f,'label'};
	}

	if ( ! defined $field{$f,'sameline'} ) {
	    print "." x ($max - length($field{$f,'label'})) ;
	    #---------------------------------
	    # ADDED: Mark required fields with a !
	    #---------------------------------
	    if ( $field{$f,'non_null'}) {
		print "!";
	    } else {
	        print":";
	    }
	}
	print "</b>";

	#--------------------------------------------------
	# Query field : Selection box or Text input field.
	#--------------------------------------------------
	if ( $field{$f,'enum'} ) {
	    print "<SELECT name=\"$f\"";
	    print " MULTIPLE" if ( defined $field{$f,'multiple'});
	    print " SIZE=$field{$f,'size'}" 
	       if (defined $field{$f,'size'});
	    print ">";
	    if (defined $field{$f,'enum_perl'}) {
	       $enum_values = eval $field{$f,'enum_perl'};
               if ( $@ ) {
                   &CgiError("Error in $f"."'s 'enum_perl' attribute:<p> $@\n");
	           goto BAIL;
               }
	    } else {
	       $enum_values = $field{$f,'enum'};
	    }
	    foreach ( split(',',$enum_values) ) {
		( $value, $title ) = /^\s*(.*)=(.*)$/;
		next if $value eq '%' && defined $field{$f,'multiple'};
		print "<OPTION value=\"$value\"> $title\n";
	    }
	    print "</SELECT>";
	#--------------
# line 850 WDB
	# Generate a system date for insertion into the database.
	} elsif ( $field{$f, 'compute_date'} ) {
             ( $sec, $min, $hour, $mday, $mon, $year, $wday, $yday, $isdst ) =
               localtime(time); 
             $mon++;
	     $mon = sprintf("%02d", $mon);
	     $mday = sprintf("%02d", $mday);
	     $year = sprintf("%02d", $year);
             $time_stamp = "$mon/$mday/$year";
	     print " <b>$time_stamp</b>";
             print "<INPUT name=\"$f\" type=hidden value=\"$time_stamp\" >";
	#--------------
	# Generate a textbox to enter large amounts of data.
	#--------------
        } elsif ( defined $field{$f, 'textarea'} ) {
	   print "<TEXTAREA NAME=\"$f\"";
	   print " $field{$f, 'textarea'}" if $field{$f, 'textarea'} ne "TRUE";
	   print ">";
	   if ( $field{$f,'default'} ) {
	      $default = eval $field{$f,'default'};
	      if ( $@ ) {
                 &CgiError("Error in $f"."'s 'default' attribute:<p>\n$@\n");
		 &dbi_disconnect($db_query, \$main_th);
	         goto BAIL;
	      }
	   } else {
	      $default = "";
	   }
           $default =~ s/^\s*//;
           $default =~ s/\s*$//;
	   print $default;
	   print "</TEXTAREA>";
	} else {
	    if ( $field{$f,'length'} > (55-$max) ) {
		$length = (55-$max) ;
	    } else {
		$length = $field{$f,'length'};
	    }
	    if ( $field{$f,'default'} ) {
		$default = eval $field{$f,'default'};
		if ( $@ ) {
                    &CgiError("Error in $f"."'s 'default' attribute:<p>\n$@\n");
	            goto BAIL;
		}
	    } else {
		$default = "";
	    }
	    print "<INPUT name=\"$f\" size=$length ";
	    print "MAXLENGTH=$field{$f,'max_length'} " 
# line 900 WDB
	       if defined $field{$f,'max_length'};
	    print "value=\"$default\" >";
	}
	#-------------
	# Unit label
	#-------------
	print " $field{$f,'unitlabel'}" if ( $field{$f,'unitlabel'} );
    }

    #------------------------------
    # Print the rest of the page.
    #------------------------------
    print "</pre> ";

    &write_menu;

    print "</FORM>\n";
    &html_tail unless defined $form{'NO_TAIL'};
    print "</BODY></HTML>\n";
}
#
#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#.PURPOSE	Print an HTML form for an update or delete.  Data from a
#                previous select statement is displayed in the form for the
#                user to update, or verify the information before a delete.
#
#.RETURNS	True, or not at all ....
#------------------------------------------------------------------------------
sub PrintProcessingForm 
{
    my( $f, $max, $value, $title, $temp, $rows, $length, %firstrow, $keyval );
    my( $time_stamp, $help, $numrows, $key_value, $notfirstfield, $err );
    my( $sec, $min, $hour, $mday, $mon, $year, $wday, $yday, $isdst, $default );
    my ($enum_values);
    $default = "";

    #-----------------------------
    # Print HTML header stuff
    #-----------------------------
    &write_header;

    if (@fields < 1) { 
       print "<B>ERROR:</B> There are no fields to ";
       print "update!" if $ACTION eq "update_form";
       print "delete!" if $ACTION eq "delete_form";
       print "<BR>\nPlease check the FDF syntax and/or your query.<BR>\n";
       goto BAIL; 
    }

# line 950 WDB
    if (!$KEY_ARGS) {
       print "<B>ERROR:</B> No record has been selected for ";
       print "update!" if $ACTION eq "update_form";
       print "deletion!" if $ACTION eq "delete_form";
       print "<BR>You must run a database query and select a record from the ".
	     "query results form before calling this form.<BR>\n";
       goto BAIL;
    } else {
	foreach $keyval ( split( '::', $KEY_ARGS ) ) {
	    print "\nFound key: $keyval\n" if $debug;
	    $in{pop(@keys)} = $keyval;
	}
    }

    &write_menu;

    return if !&ExecuteQuery;

    #----------------
    # Print all rows.
    #----------------

    $numrows=0;
    if ( ! (%firstrow = &dbi_nextrow($main_th, @select_titles) ) ) {
        print "<p><h2>No data returned !</h2>\n";
        print "Please go 'Back' to the query form and re-specify your query.\n";
        print "<p><hr>\n";
        print "<h3>Your Query :</h3>\n";
        foreach $f ( @fields ) {
            if ( defined $in{$f} && $in{$f} ne "" ) {
                print "<em>$field{$f,'label'}</em> : $in{$f}<br>\n";
            }
        }
        print "<p><hr>\n";
        print "<h3>SQL Query used.</h3>\n";
        print "<em>select $selectlist</em> \n";
        print "<em>from</em> $fromlist<br>\n";
        print "$wherelist <br>\n";
        &html_tail unless defined $form{'NO_TAIL'};
        print "</BODY></HTML>\n";
        return;
    }
    $numrows++;

    print "\n<pre>";

    #------------------------
    # Print fields
    #------------------------
# line 1000 WDB
    $max = 0;

    #---------------------------------
    # This finds the longest field for formatting purposes.
    #---------------------------------
    foreach $f ( @fields ) {
       #next if defined $field{$f,'serial_field'} && $ACTION eq "update_form";
       next if defined $field{$f,'hidden'};
       $max = length($field{$f,'label'})+1 
          if length($field{$f,'label'})+1 > $max;
    }

    foreach $f ( @fields ) {
        next if defined $field{$f,'hidden'};

	#---------------
	# Serial key fields (those which are unique identifiers for a row in
	#  the database) should not be updateable by the user, but it is
	#  imperative they get passed to the DoUpdate() routine.  Any field
	#  with the 'no_update' parameter is included in the Web form as a
	#  hidden element to be sure they don't change in the database during
	#  the update.
	#---------------
	if ($ACTION eq "update_form" && defined $field{$f, 'no_update'} 
	    && !defined $field{$f, 'computed'}){
	    $key_value = &Value( $f, \$firstrow{$field{$f,'column'}}, 
		                   \%firstrow, \%field );
            $key_value =~ s/^\s*//;
            $key_value =~ s/\s*$//;
	    print "<INPUT name=\"$f\" type=hidden value=\"$key_value\">";
	    print "<INPUT name=\"old_value_$f\" type=hidden ".
	          "value=\"$key_value\">";
	    next;
        }

	if (defined $field{$f,'form_buttons'}) {
	   &write_menu;
	} else {
	   print "\n" if ( $notfirstfield++ && !defined $field{$f,'sameline'} );
	}
	print &eval_attr( $f,'html') if defined $field{$f,'html'};

	#---------------------------------
	# Label
	#---------------------------------
	print "<b>";
	if ( $field{$f,'help'} ) {
	    print "help field : $field{$f,'help'}<BR>\n" if $debug;
            $help = eval $field{$f,'help'};
            if ( $@ ) {
               &CgiError("Error in $f"."'s  'help' attribute :<p>\n $@\n");
               &dbi_disconnect($db_query, \$main_th);
               goto BAIL;
# line 1050 WDB
            }
	    print "<a href=\"";
	    print $form{'DOCURL'} if ( $help =~ /^#.*/ );
	    print $help . "\">";
	    print $field{$f,'label'};
	    print "</a>";
	} else {
	    print $field{$f,'label'};
	}

	if ( ! defined $field{$f,'sameline'} ) {
	    print "." x ($max - length($field{$f,'label'})) ;
	    #------------------------------
	    # Mark required fields with a !
	    #------------------------------
	    if ( $ACTION eq "update_form" && $field{$f,'non_null'} ) {
		print "!";
	    } else {
	        print":";
	    }
	}
	print "</b>";

	# Pick the output format to display the data in.
	if ( defined $field{$f,'enum'} ) {
	    print "<SELECT name=\"$f\">";
	    $temp = &Value( $f, \$firstrow{$f}, \%firstrow, \%field );
            $temp =~ s/^\s*//;
            $temp =~ s/\s*$//;
	    if (defined $field{$f,'enum_perl'}) {
	       $enum_values = eval $field{$f,'enum_perl'};
               if ( $@ ) {
                   &CgiError("Error in $f"."'s 'enum_perl' attribute:<p> $@\n");
	           goto BAIL;
               }
	    } else {
	       $enum_values = $field{$f,'enum'};
	    }
	    foreach ( split(',',$enum_values) ) {
		( $value, $title ) = /^\s*(.*)=(.*)$/;
		#print "Title: ($title), Temp: ($temp)<BR>\n";
		if ( $value eq $temp ) {
		    print "<OPTION SELECTED value=\"$value\"> $title\n";
		} else {
		    print "<OPTION value=\"$value\"> $title\n";
		}
	    }
	    print "</SELECT>";
	#--------------
# line 1100 WDB
	# Generate a system date for insertion into the database.  When a
	#  record is updated, it gets the new system date.  If you don't want
	#  this to happen, comment out this code, or figure out a new field
	#  parameter to prevent it, like no_compute_date_on_update.  Put the
	#  code here, test it, and send it to me when it works. :)
	#--------------
        } elsif ( defined $field{$f, 'compute_date'} 
	          && $ACTION ne "delete_form") {
             ( $sec, $min, $hour, $mday, $mon, $year, $wday, $yday, $isdst ) =
               localtime(time);
             $mon++;
	     $mon = sprintf("%02d", $mon);
	     $mday = sprintf("%02d", $mday);
	     $year = sprintf("%02d", $year);
             $time_stamp = "$mon/$mday/$year";
	     if (defined $field{$f,'to_db'}) {
	        $time_stamp = eval $field{$f,'to_db'};
	        if ( $@ ) {
	           $err =  "Error in $f"."'s 'to_db' attribute :";
	           $err .=  "<p>\n $@\nPlease report this error.\n";
	           &CgiError($err);
	           goto BAIL;
	        }
	     }
             print " <b>$time_stamp</b>";
             print "<INPUT name=\"$f\" type=hidden value=\"$time_stamp\" >";
	#--------------
	# Generate a textbox to enter large amounts of data.
	#--------------
        } elsif ( defined $field{$f, 'textarea'} ) {
	   print "<TEXTAREA NAME=\"$f\"";
	   print " $field{$f, 'textarea'}" if $field{$f, 'textarea'} ne "TRUE";
	   print ">";
	   if ( $field{$f,'default'} ) {
	      $default = eval $field{$f,'default'};
	      if ( $@ ) {
                 &CgiError("Error in $f"."'s 'default' attribute:<p>\n$@\n");
		 &dbi_disconnect($db_query, \$main_th);
	         goto BAIL;
	      }
	   } else {
	      $default = "";
	      $default = &Value( $f, \$firstrow{$f}, \%firstrow, \%field );
	   }
           $default =~ s/^\s*//;
           $default =~ s/\s*$//;
	   print $default;
	   print "</TEXTAREA>";
	} else {
# line 1150 WDB
	    if ( $field{$f,'length'} > (55-$max) ) {
		$length = (55-$max) ;
	    } else {
		$length = $field{$f,'length'};
	    }
	    if ( $field{$f,'default'} ) {
		$default = eval $field{$f,'default'};
		if ( $@ ) {
                    &CgiError("Error in $f"."'s 'default' attribute:<p>\n$@\n");
		    &dbi_disconnect($db_query, \$main_th);
	            goto BAIL;
		}
	    } else {
		$default = &Value( $f, \$firstrow{$field{$f,'column'}}, 
		                   \%firstrow, \%field );
                $default =~ s/^\s*//;
                $default =~ s/\s*$//;
	    }
	    print "<INPUT name=\"$f\" size=$length ";
	    print "MAXLENGTH=$field{$f,'max_length'} " 
	       if defined $field{$f,'max_length'};
	    print "value=\"$default\" >";
	}
	#-------------
	# Unit label
	#-------------
	print " $field{$f,'unitlabel'}" if ( $field{$f,'unitlabel'} );

	# Need the old values to make sure we match the proper record when
	#  updating.  i.e. 'update where f1 = v1 and f2 = v2', etc.
	$default = $firstrow{$field{$f,'column'}} 
	   if defined $firstrow{$field{$f,'column'}};
	$default = &dbi_get_null_assignment if $default eq "";
	print "<INPUT name=\"old_value_$f\" type=hidden value=\"$default\">";

    }

    &dbi_disconnect($db_query, \$main_th);

    #------------------------------
    # Print the rest of the page.
    #------------------------------
    print "</pre> ";

    &write_menu;

    print "</FORM>\n";
    &html_tail unless defined $form{'NO_TAIL'};
    print "</BODY></HTML>\n";
# line 1200 WDB
}
#
#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#.PURPOSE	Retrieve a query, execute it, and print the result.
#------------------------------------------------------------------------------
sub DoQuery 
{
    my ($startcount, $endcount, $pagecount, $count_result, $dummy, $thing, $f);
    my ($numrows, $formvars, %myrow, %firstrow, $stuff, $one, $two, $countrows);
    my ($lastrow);

    #--------------------------
    # Print HTML Header stuff.
    #--------------------------
    &write_header;

    return if !&ExecuteQuery;

    #----------------
    # Print all rows.
    #----------------

    #----------------
    # This whole row paging thing got started when someone wanted the number
    #  of records displayed at the top of the results form, instead of at the
    #  bottom.  Now WDB can display a page of data at a time, under one
    #  condition: there must be a single, unique field used as the paging key,
    #  and the data must come out of the database ordered by that paging key
    #  field. This can be done with an 'order by' clause, or by clustering an
    #  index on that field in the database.
    #----------------
    $startcount = $endcount = $pagecount = $countrows = $lastrow = 0;

    #----------------
    # $sql_count is a SQL statement composed in the ExecuteQuery() function at
    #  the same time the regular query is constructed.  Mini SQL just needs
    #  a cursor handle of the query itself to find the number of rows.  Other
    #  databases must issue a "select count(*)" SQL statement.  mSQL 1.0.x
    #  doesn't support count(*) so it uses @select_titles.
    #----------------
    $count_result = &dbi_countrows($db_query, $sql_count, @select_titles);

    #----------------
    # If the total_rows element exists, this is a request for a page of
    #  results beyond the first page.  This piece of code keeps track of where
    #  you are in the total result set.
    #----------------
    if ( defined $in{'total_rows'} && $in{'total_rows'} != 0 ) {
        $countrows = $in{'total_rows'};
# line 1250 WDB
	$count_result -= $in{'lastrow'} if defined $in{'lastrow'};
        $startcount = $countrows - $count_result + 1;
        $endcount = $startcount + $rowcount - 1;
        $endcount = $in{'total_rows'} if $in{'total_rows'} < $endcount;
        $pagecount = "$startcount - $endcount";
    } else {
        $countrows = $count_result;
        $startcount = 1;
        $endcount = $startcount + $rowcount - 1;
        $pagecount = "$startcount - $endcount";
    }

    #------------
    # Here the code reporting the number of matches from a query is reported
    #  at the top of the results page, instead of at the bottom.  If the
    #  ROW_PAGING parameter is specified in the FDF, results are paged.
    #------------
    if ( $countrows <= $rowcount ) {
        print "<p><b>Matches: $countrows.</b>\n";
        if ( defined $in{'total_rows'} && $in{'total_rows'} ne "" ) {
            print "<b>$pagecount are shown on this page.</b>\n";
        }
    } elsif ( defined $form{'ROW_PAGING'} ) {
        print "<p><b>Matches: $countrows. ";
        print "$pagecount are shown on this page.</b>\n";
        if ( !defined $in{'total_rows'} || $in{'total_rows'} > $endcount ) {
            print "<b>Click on the button at the bottom of the page for more ".
		  "records.</b>\n";
        }
    } else {
        print "<p><b>Matches: $rowcount.";
        print " Any remaining matches were ignored.</b>\n";
    }

    $numrows=0;
    if ( ! (%firstrow = &dbi_nextrow($main_th, @select_titles) ) ) {
	print "<p><h2>No data returned !</h2>\n";
	print "Please go 'Back' to the query form and re-specify your query.\n";
	print "<p><hr>\n";
	print "<h3>Your Query :</h3>\n";
	foreach $f ( @fields ) {
	    if ( defined $in{$f} && $in{$f} ne "" ) {
		print "<em>$field{$f,'label'}</em> : $in{$f}<br>\n";
	    }
	}
	print "<p><hr>\n";
	print "<h3>SQL Query used.</h3>\n";
	print "<em>select $selectlist</em> \n";
	print "<em>from</em> $fromlist<br>\n";
# line 1300 WDB
	print "$wherelist <br>\n";
	&html_tail unless defined $form{'NO_TAIL'};
        print "</BODY></HTML>\n";
	return;
    }
    $numrows++;

    #----------
    # Determine the output mode of the query results.
    #----------
    $output_mode='full';
    $output_mode='tab' unless defined $in{'full_screen_mode'};
    $output_mode='full' if defined $form{'FORCE_FULL'};
    $output_mode='tab' 
      if $ACTION eq "update_query" or $ACTION eq "delete_query";

    &write_menu;

    if ( %myrow = &dbi_nextrow($main_th, @select_titles)  ) {

	if (!defined $in{'lastrow'}) {
	    if ( $output_mode eq 'full' ) {
	        &PrintVars(%firstrow);
	    } else {
	        &PrintTabHead;
	        print "<TR>\n" if defined $form{'TABLE_RESULTS'};
	        &PrintTabVars(%firstrow);
	        print "</TR>\n" if defined $form{'TABLE_RESULTS'};
	    }
            $numrows=1;
	} else {
	    $numrows = 1;
	    while ($numrows < $in{'lastrow'}) {
	        $numrows++;
	        %myrow = &dbi_nextrow($main_th, @select_titles);
	    }
	    &PrintTabHead if $output_mode ne 'full';
            $numrows=0;
	}
	goto SINGLE_RECORD if $rowcount < 2 && $numrows == 1;
	do {
	    $numrows++;
	    if ( $output_mode eq 'full' ) {
		&PrintVars(%myrow);
	    } else {
	        print "<TR>\n" if defined $form{'TABLE_RESULTS'};
		&PrintTabVars(%myrow);
	        print "</TR>\n" if defined $form{'TABLE_RESULTS'};
	    }
# line 1350 WDB
	} while( $numrows < $rowcount 
		 && (%myrow = &dbi_nextrow($main_th, @select_titles)) );
SINGLE_RECORD:
        if ( $output_mode ne 'full' ) {
            if ( $form{'TABLE_RESULTS'} ) {
                  print "</TABLE>\n";
            } else {
                print "</pre>\n";
            }
        }
	#------------
        # ROW_PAGING is set in the FDF file
	#------------
        if ( defined $form{'ROW_PAGING'}
             && ( ($in{'total_rows'} && $in{'total_rows'} > $endcount)
             || (!$in{'total_rows'} && $numrows < $countrows)) ) {
	   print "<form action=\"$ENV{'SCRIPT_NAME'}$ENV{'PATH_INFO'}\" ".
	         "method=post>\n";
           foreach $thing ( @in ) {
             ( $one, $two ) = (split(/<>|<=|>=|!=|<|>|==|=|~/, $thing));
             $two =~ s/\+/ /g;
             $two =~ s/%(..)/pack("c",hex($1))/ge;
	     print "   <input type=hidden name=\"$one\" value=\"$two\">\n"
                if $one ne "lastrow";
           }
           print "   <input type=hidden name=\"total_rows\" ".
	         "value=\"$countrows\">\n" if !defined $in{'total_rows'};
	   $lastrow = $numrows;
	   $lastrow += $in{'lastrow'} if defined $in{'lastrow'};
	   print "   <input type=hidden name=\"lastrow\" value=\"$lastrow\">\n";
	   print "   <input type=submit value=\"Continue to Next Page of ".
	         "Results\">\n";
	   print "</form>\n";
        }
    } else {
	#---------
	# Update and delete actions require tabular output even if only one
	# row is returned.  If a user forces full-screen mode, give it to
	# them, but updates and deletes are disabled.
	#---------
	if (  $output_mode eq 'full' ) {
	    &PrintVars(%firstrow);
        } else {
	    &PrintTabHead;
	    &PrintTabVars(%firstrow);
            if ( $form{'TABLE_RESULTS'} ) {
                print "</TABLE>\n";
            } else {
                print "</pre>\n";
            }
        }
    }
# line 1400 WDB

    &dbi_disconnect($db_query, \$main_th);
    &write_menu;
    &html_tail unless defined $form{'NO_TAIL'};

    if ( $debug ) {
	print "<HR><h2>Variables from form</h2>\n";
	$formvars = &PrintVariables(%in);
	print "$formvars <HR><p>\n";
	print "<h2>Query</h2>\n";
	print "use $form{'DATABASE'}<br>\ngo<br>\n";
	print "select $selectlist from $form{'TABLE'}<br>\n";
	print "$wherelist <br>\n";
	print "<hr>\n";
    }
    print "</BODY></HTML>\n";
}
#
#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#.PURPOSE	Constructs an SQL select statement and executes it.
#
#.RETURNS	Nothing, really, but leaves an open cursor with results.
#------------------------------------------------------------------------------
sub ExecuteQuery
{
    my ($key, $keyval, @tablenames, @select, $part_one, $part_two, $qual, $val);
    my ($col, $ret, $val2, $opname, $oper, @or_where, $i, $f, $val1, $dummy);
    my (@where, $err_message, $junk, $tablename, $fieldname, $prime_table);
    my ($table2, $multi_table, $count, @my_where, $save_val, $wherestuff);
    my ($wild_op, $use_parens);

    #----------------------------------------------------------
    # Get qualification either from the form or from the path
    #----------------------------------------------------------
    if ( ! $KEY_ARGS && $ACTION eq "query" ) {
        &ReadParse(\%in, \@in, \$in);
    #-----------
    # If row paging is being performed, the information passed from the
    #  original form is now being passed in the URL.  Nasty, but it works.
    #  The URL information must be decoded from HTTP format.  This code
    #  just copies what Steven Brenner's ReadParse() does.
    #-----------
    } elsif ( $ACTION eq "page" ) {
        @in = split(/[&;]/, $KEY_ARGS);
        foreach $i (0 .. $#in) {
          $in[$i] =~ s/\+/ /g;
	  ($key, $val) = split(/=/,$in[$i],2);
          $key =~ s/%(..)/pack("c",hex($1))/ge;
          $val =~ s/%(..)/pack("c",hex($1))/ge;
# line 1450 WDB
          $in{$key} .= "\0" if (defined($in{$key}));
          $in{$key} .= $val;
        }
    } elsif ($KEY_ARGS) {
	foreach $keyval ( split( '::', $KEY_ARGS ) ) {
	    print "\nFound key: $keyval\n" if $debug;
	    $in{pop(@keys)} = $keyval;
	}
        &ReadParse(\%in, \@in, \$in);
    } else {
        &ReadParse(\%in, \@in, \$in);
    }

    #--------------------------------------------------
    # Compose select and where list for the SQL select
    #--------------------------------------------------
    @tablenames = @or_where = @my_where = ();
    $wherestuff = "";

    # Used for sub query trickery.
    $multi_table = 0;
    ($prime_table, $table2) = (split('\,', $form{'TABLE'}));
    # Take off leading and trailing blanks.
    $prime_table =~ s/^\s*//;
    $prime_table =~ s/\s*$//;
    foreach $f ( @fields ) {

	if ( $field{$f,'column'} ) {
	    $col = $field{$f,'column'};
	} else {
	    $col = $f;
	}

       # Used for sub-select trickery.
       unless ( defined $field{$f,'ignore_select'} ) {

             # This sets up processing for multi-value attributes.
	     if (!$multi_table) {
                ($tablename, $fieldname) = (split('\.', $col ));
                if ( defined $fieldname && $fieldname ne "" ) {
                  $tablename =~ s/^\s*//;
                  $tablename =~ s/\s*$//;
                  if ( $tablename ne $prime_table ) {
                    $multi_table = 1;
                  }
                }
	     }

	      unless ( defined $field{$f,'computed'} ) {
	         push( @select, ( $field{$f,'column'} ) 
	   		         ? $field{$f,'column'} : $f ); 
# line 1500 WDB
	         push( @select_titles, ($field{$f,'column'} 
		         ? $field{$f,'column'} : $f) );
	      }
       }

	if ( $ACTION eq "default" ) { 
	    if ( $field{$f,'default'} && ! $in{$f} ) {
		$in{$f} = &SGMLtoASCII( &eval_attr($f,'default',0) );
	    }
	}

	if ( defined $in{$f} && $in{$f} ne "" ) {
	  next if $in{$f} eq '%';       # Ignore '%' fields (from enum fields)

          # This sets up processing for multi-value attributes.
	  if (!$multi_table) {
             ($tablename, $fieldname) = (split('\.', $col ));
             if ( defined $fieldname && $fieldname ne "" ) {
                $tablename =~ s/^\s*//;
                $tablename =~ s/\s*$//;
                if ( $tablename ne $prime_table ) {
                   $multi_table = 1;
                }
             }
	  }

	  if ( defined $field{$f,'computed'} ) { 
	      print "Evaluating computed field : $f = $in{$f}<BR>\n" if $debug;
	      print "- Compute : $field{$f,'to_db'}<BR>\n" if $debug;
	      $ret = &eval_attr($f,'to_db',$in{$f});
	      print " = $ret<BR>\n" if $debug;
	      next;	# Don't query on computed fields.
	  }

	  print "Parsing $f = '$in{$f}'<BR>\n" if $debug;

	  #------------
	  # Values input into fields on the query form are processed out here.
	  #  This line looks for 'OR' queries.
	  #------------
	  foreach $junk ( split( /\s*\|\s*|\s+OR\s+|\s+or\s+/, $in{$f} ) ) {

	    # This looks for multiple values for a single field, like a
	    #  list that allows multiple selections.
	    foreach $qual (split("\0", $junk)) {

	    print "      qual = '$qual'<BR>\n" if $debug;

	    #------------
# line 1550 WDB
	    # This line breaks off operators from the query parameter.  I.e.,
	    #  if the user enters '>1024', $oper = > and $val = 1024.
	    #------------
	    ( $oper, $val ) = $qual =~ /^\s*(<>|<=|>=|!=|<|>|==|=|~)?(.*)$/;

	    #------------
	    # This upper cases all query parameters to make database searches
	    #  more efficient than doing a search on all possible combinations
	    #  of cases.  Of course, this assumes all your data is in uppercase
	    #  in the database.  That's an important point. :)
	    #------------
	    $val =~ tr/a-z/A-Z/ if defined $form{'UPCASE_QUERY'};

	    $val2 = "";
	    # Get range - (1..10) etc.
	    ( $val1, $val2 ) = split( /\.\./,$qual,2 ) unless $oper;                    
            if ( !$oper && $field{$f, 'operators'} ) {
                $opname = $f."_op";
                $oper = $in{$opname};
            } elsif ( !$oper && $field{$f, 'enum'} ) {
                $oper = "==";
            }
	    if ($oper && ($oper eq "!=" || $oper eq "<>")) {
	       $oper = &dbi_get_not_equal;
	    }

	    if ( $val2 ) {
		print "   range: '$val1' -> '$val2'<BR>\n" if $debug;

		if (   $field{$f,'type'} eq 'char' 
		    || $field{$f,'type'} eq 'datetime' ) {
		    push( @or_where, "$col between '"
				     . &eval_attr($f,'to_db',$val1)
				     . "' and '"
				     . &eval_attr($f,'to_db',$val2) . "'");
		} else {
		    push( @or_where, 
			  "$col between ". &eval_attr($f,'to_db',$val1)
			  . " and " . &eval_attr($f,'to_db',$val2) );
		}
	    } elsif ( defined $val ) {
		$val1 = $val;
		if (defined $field{$f,'to_db'}) {
		   $val = eval $field{$f,'to_db'};
		}
	        if ( $@ ) {
	            $err_message =  "Error in $f"."'s 'to_db' attribute :";
	            $err_message .=  "<p>\n $@\nPlease report this error.\n";
	            &CgiError($err_message);
# line 1600 WDB
	            goto BAIL;
	        }
		if ($val eq "") { $val = $val1 };
		if ( ! $oper ) {
		    $oper = "==" if defined $KEY_ARGS;
		    $oper = "==" if defined $field{$f,'enum'};
		} 

		if ( $oper && $oper ne "==" ) {
		    $val =~ s/^\s*//;
		    $val =~ s/\s*$//;
		}
		print "      val = '$val'<BR>\n" if $debug;
	        print "     oper = '$oper'<BR>\n" if $debug;

		if ( $field{$f,'type'} eq 'char' ) {
		    $save_val = $val;
		    print "type: char<BR>\n" if $debug;
		    #--------
		    # You might need to comment one or both of these out,
		    #  depending on your database.
		    #--------
		    $val = &dbi_escape_characters($val);

		    #-----------
		    # This section creates a wildcard matching string that will
		    #  match data in any possible upper or lowercase 
		    #  combination.  This is one of the slowest database
		    #  queries you can do, but it's nice to have.
		    #-----------
		    if ( ! $oper || $oper eq "=" || $oper eq "~" ) { # Default 
			$oper = &dbi_get_stringmatch_operator;
			$val = &dbi_create_case_insensitive_string($val)
			   unless defined $form{'UPCASE_QUERY'};
			$wild_op = &dbi_get_wildcard_operator;
			$val  = $wild_op.$val.$wild_op;
		    } elsif ( $oper eq "==" ) {
			$oper = "=";
		    }
		    push( @or_where, "$col $oper '$val'");
		    push( @my_where, "<b>$field{$f,'label'}</b> ".
				     "$oper <b>$save_val</b>");

		} elsif ( $field{$f,'type'} eq 'datetime'  ) {
		    $oper = &dbi_get_stringmatch_operator unless $oper;
		    $oper = "=" if $oper eq "==";
		    push( @my_where, 
			  "<b>$field{$f,'label'}</b> $oper <b>$val</b>");
		    push( @or_where, "$col $oper '$val'");
# line 1650 WDB

		} else {
		    if ($oper eq "~") { $oper = "="; }
		    $oper = "=" unless $oper;       # Default operand is '='
		    $oper = "=" if $oper eq "==";
		    push( @my_where, 
			  "<b>$field{$f,'label'}</b> $oper <b>$val</b>");
		    push( @or_where, "$col $oper $val");
		}
	    }
	  }}
	  print "     where = ".join(" OR ",@or_where)."<BR>\n" if $debug;

	  if ( @or_where > 1 ) {
	    if (&dbi_can_use_parentheses) {
	       push( @where, "(". join(" OR ",@or_where) . ")" );
	    } else {
	       push( @where, join(" OR ",@or_where) );
	    }
	  } else {
	    push( @where, @or_where );
	    if (!@my_where) {
	       push( @my_where, "<b>$field{$f,'label'}</b> $oper <b>$val</b>");
	    }
	  }
	  @or_where = ();
       }
    }
    #------------
    # Get query constraints and join conditions from the FDF and add them to
    #  the query.
    #------------
    push( @where, $form{'CONSTRAINTS'}) if defined $form{'CONSTRAINTS'};
    push( @where, $form{'JOIN'}) if defined $form{'JOIN'} && $multi_table == 1;

    $selectlist = join(',',@select);
    $wherelist  = join(' AND ', @where );
    # Added for multi-value attributes
    #
    if ( $multi_table == 1 ) {
      $fromlist = $form{'TABLE'};
    } else {
      ($prime_table, $table2) = (split('\,', $form{'TABLE'}));
      $fromlist = $prime_table;
    }

    # This part "pretties up" the search criteria so non-database people
    #  can understand it.
    if ( $wherelist ) {
# line 1700 WDB
        $wherestuff = join( ' <BR>', @my_where );
        $wherestuff =~ s/matches/contains/g;
        $wherestuff =~ s/==/equals/g;
        $wherestuff =~ s/=/equals/g;
    }

    #------------
    # This stuff is for counting records returned from a query so the number
    #  of matches can be reported at the top of the results page instead of
    #  at the bottom.
    #------------
    $count_wherelist = $wherelist = "where ".$wherelist if $wherelist;
    $sql_count = "select count(*) from $fromlist $wherelist";

    $wherelist .= " order by $form{'ORDER'}" if defined $form{'ORDER'};

    #-------------------------------
    # Open database and select data.
    #-------------------------------
    $result = "";
    $result = 
       &dbi_connect( $GLOBAL_USER, $GLOBAL_PSWD, $GLOBAL_SERVER, 
                     $form{'DATABASE'}, \$db_query);
       
    if ($result) {
       &dbi_disconnect($db_query, \$main_th);
       print $result;
       goto BAIL;
    }

    #-------------------------------
    # If wanted, show the query parameters entered by the user at the top of
    #  the results page.
    #-------------------------------
    if (defined $form{'SHOW_QUERY'} ) {
        print "<h3>Search Criteria :</h3>\n";
        if ( $wherestuff ) {
            print "$wherestuff";
        } else {
            print "<b>None.</b>\n";
        }
    }

    if ( defined $in{'purpose'} && $in{'purpose'} eq "count" ) {
       $count = &dbi_countrows($db_query, $sql_count, @select_titles);
       print "<p><b> Your query matched: $count records.</b>\n";
       &dbi_disconnect($db_query, \$main_th);
       goto BAIL;
    }
# line 1750 WDB

    if ( $in{'max_rows_returned'} ) {
	$rowcount = $in{'max_rows_returned'};
	$rowcount = $GLOBAL_MAXROWCOUNT if $rowcount > $GLOBAL_MAXROWCOUNT;
    }
    &dbi_rowcount( $rowcount );
    &dbi_dateformat( $DATEFORMAT ) if ( $DATEFORMAT );

#$debug = 1;
    print "<BR>\n**** QUERY ****<BR><BR>\n" if $debug;
    print "use $form{'DATABASE'}<BR><BR>\n" if $debug;
    print "select $selectlist from $fromlist<BR> <BR>\n" if $debug;
    print "$wherelist<BR> \n<BR>\n" if $debug;
#$debug = 0;
    $result = "";
    $result = &dbi_dosql($db_query, \$main_th, 
	       "select $selectlist from $fromlist $wherelist");

    if ($result) {
       print $result;
       return 0;
    } else {
       return 1;
    }
}
#
#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#.PURPOSE	Retrieve an insert, execute it, and print the result.
#
#.REMARKS       Also works across join conditions, under special circumstances.
#
#.RETURNS	A serial key value, if required.
#------------------------------------------------------------------------------
sub DoInsert 
{
    my ($keyval, $die, $die2, $f, $table_name, $field_name, $m, $value_list);
    my (@join_fields, $err, @null_fields, $primary, $primary_key, $ret, $val);
    my (@insert_values, @insert_fields, $data, @tables, $t, %foreign_keys);
    my (@insert_join_fields, @insert_join_values, $counter, $insert_into);
    my ($post_insert, $post_insert_table, $post_insert_field, @temp_tables);

    #--------------------------
    # Print HTML Header stuff.
    #--------------------------
    &write_header;
    &write_menu;

    &ReadParse(\%in, \@in, \$in);

# line 1800 WDB
    #------------------------------------------------------------------------
    # First make sure all required fields are filled.
    # If this is a join form, make sure all field names are fully qualified.
    # I don't care if the database will support non-qualified fields, because
    # the qualified field names are the only way I know which table to insert
    # the data into.
    #------------------------------------------------------------------------
    $die = $die2 = $primary = "";
    $post_insert = $post_insert_table = $post_insert_field = "";
    %foreign_keys = ();

    foreach $f ( @fields ) {

        if ( defined $field{$f, 'non_null'} && ( !defined $in{$f} ||
	     ( defined $in{$f} && $in{$f} eq "" )) ) {
            if ( $die ne "TRUE" ) {
		$die = "TRUE";	
	    }
	    push( @null_fields, "$f" );
        }

        # This is a kludge to ignore the 'MORE' field.
	if ( $form{'JOIN'} && defined $field{$f, 'column'} ) {
	    ($table_name, $field_name) = split( /\./, $field{$f, 'column'});
	    if ( ! $field_name ) {
		$die2 = "TRUE";
		push( @join_fields, "$f" );
	    }
	}
    }

    #--------------------------------------------------
    # If a required field is empty, or a field isn't qualified, say so.
    #--------------------------------------------------
    if ( $die ) {
	$err = "Required field(s): ".join(',',@null_fields)." cannot be empty.";
        print "$err<br>";
	return;
    }
    if ($die2) {
	$err = "field(s): ".join(',',@join_fields)." must be fully qualified";
        print "On an insert, $err in the FDF.<br>";
	return;
    }

    #-------------------------------
    # Open the database here so WDB can get a primary key value if needed.
    #-------------------------------
    $result = "";
# line 1850 WDB
    $result = 
       &dbi_connect( $GLOBAL_USER, $GLOBAL_PSWD, $GLOBAL_SERVER, 
                     $form{'DATABASE'}, \$db_insert );

    if ($result) {
       print $result;
       &dbi_disconnect($db_insert, \$insert_th);
       goto BAIL;
    }

    #--------------------------------------------------
    # Compose insert into and values list for the SQL statement.
    #--------------------------------------------------
    foreach $f ( @fields ) {

	if ( defined $field{$f,'computed'} ) { 
	   print "Evaluating computed field : $f = $in{$f}<BR>\n" if $debug;
	   print "- Compute : $field{$f,'to_db'}<BR>\n" if $debug;
	   $ret = eval $field{$f,'to_db'};
           if ( $@ ) {
              $err = "Error in $f"."'s 'to_db' attribute :".
                     "<p>\n $@\nPlease report this error.\n";
              &CgiError($err);
              &dbi_disconnect($db_insert, \$insert_th);
              goto BAIL;
           }
	   print " = $ret<BR>\n" if $debug;
	   next;	# Don't insert into computed fields.
	}

	$val = "";
	$val = $in{$f} if defined $in{$f};

        #-------------------------------------------------------------------
	# Take care of primary key values, then foreign key values, then the
	# other values.  If there are foreign keys in a join, the primary key 
	# they refer to must come first in the FDF file.
	# Note: A serial field is one which is self-incrementing on an insert.
	#       A foreign key is one which refers to a primary key in another
	#       table.  If a foreign key references a serial key, the serial
	#       key value must be inserted first, then returned so it can be
	#       used as the foreign key in other tables.
	#-------------------------------------------------------------------
        if ( defined $field{$f, 'primary_key'} ) {

	    $primary = 1;
	    if ( defined $field{$f, 'serial_field'} ) {
	       if ($field{$f, 'serial_field'} eq "pre_insert") {
	           if ( $form{'JOIN'} ) {
# line 1900 WDB
	               ($table_name, $field_name) 
		          = split( /\./, $field{$f, 'column'});
	           } else {
	               $table_name = "$form{'TABLE'}";
	           }
	           $primary_key 
		       = &dbi_getnextserialvalue($db_insert, "$table_name");
		   $val = $primary_key;
                   if (defined $field{$f,'to_db'}) {
                      $primary_key = eval $field{$f,'to_db'};
                      if ( $@ ) {
                         $err = "Error in $f"."'s 'to_db' attribute :".
                                "<p>\n $@\nPlease report this error.\n";
                         &CgiError($err);
                         &dbi_disconnect($db_insert, \$insert_th);
                         goto BAIL;
                      }
                   } 
	       } elsif ($field{$f, 'serial_field'} eq "post_insert") {
		   $post_insert = "empty";
	           if ( $form{'JOIN'} ) {
	               ($post_insert_table, $post_insert_field) 
		          = split( /\./, $field{$f, 'column'});
	           } else {
	               $post_insert_table = "$form{'TABLE'}";
		       $post_insert_field = "$field{$f, 'column'}";
	           }
	           $primary_key = &dbi_get_serial_field_placeholder;
		   print "post_insert_table: $post_insert_table, ".
		         "post_insert_field: $post_insert_field<BR>\n" if $debug;
	       } else {
	          &CgiError("Unrecognized type of serial field for field:".
		            "$f.");
                  &dbi_disconnect($db_insert, \$insert_th);
	          goto BAIL;
	       }
	    } else {
               if (defined $field{$f,'to_db'}) {
                  $primary_key = eval $field{$f,'to_db'};
                  if ( $@ ) {
                     $err = "Error in $f"."'s 'to_db' attribute :".
                            "<p>\n $@\nPlease report this error.\n";
                     &CgiError($err);
                     &dbi_disconnect($db_insert, \$insert_th);
                     goto BAIL;
                  }
               } else {
	          $primary_key = $in{$f};
	       }
# line 1950 WDB
	    }
	    if ($primary_key eq "") {
	       &CgiError("Cannot get primary key value for field: $f.");
               &dbi_disconnect($db_insert, \$insert_th);
	       goto BAIL;
	    }
	    push( @insert_fields, ( $field{$f,'column'} )
		?  $field{$f,'column'} : $f );
	    if ( $field{$f, 'type'} eq "char" ) {
               push( @insert_values, "'$primary_key'" );
	    } else {
               push( @insert_values, "$primary_key" );
	    }

        } elsif ( defined $field{$f, 'foreign_key'} ) {
	    
	    if ( $primary ) {
		# Save the name of the foreign key field, indexed to the name
		#  of the 'column' attribute, to use later.
		$foreign_keys{$field{$f,'column'}} = $f;

	        push( @insert_fields, ( $field{$f,'column'} )

		      ?  $field{$f,'column'} : $f );
	        push( @insert_values, "$primary_key" );
	    } else {
                &dbi_disconnect($db_insert, \$insert_th);
		&CgiError("Foreign key: $f refers to a non-existant primary ".
			  "key.<br>");
                &dbi_disconnect($db_insert, \$insert_th);
		goto BAIL;
	    }

        } elsif ( defined $field{$f, 'serial_field'} ) {

           # Only pre_insert serial fields need any special handling by WDB.
	   if ($field{$f, 'serial_field'} eq "pre_insert") {
	      if ( $form{'JOIN'} ) {
	         ($table_name, $field_name) 
	            = split( /\./, $field{$f, 'column'});
	      } else {
	         $table_name = "$form{'TABLE'}";
	      }
	      $data = &dbi_getnextserialvalue($db_insert, "$table_name");
              if (defined $field{$f,'to_db'}) {
                 $data = eval $field{$f,'to_db'};
                 if ( $@ ) {
                    $err = "Error in $f"."'s 'to_db' attribute :".
                           "<p>\n $@\nPlease report this error.\n";
# line 2000 WDB
                    &CgiError($err);
                    &dbi_disconnect($db_insert, \$insert_th);
                    goto BAIL;
                 }
              }
	      push( @insert_fields, ( $field{$f,'column'} )
	         ?  $field{$f,'column'} : $f );
	      push( @insert_values, $data );
	   }

	} else { # Take care of all the other fields.

	    unless ( defined $field{$f,'computed'} ) {
		$data = "";
                if ( defined $in{$f} ) {
                    $data = $in{$f};
                }
	        push( @insert_fields, ( $field{$f,'column'} ) 
		        ?  $field{$f,'column'} : $f ); 

                if (defined $field{$f,'to_db'}) {
                   $data = eval $field{$f,'to_db'};
                   if ( $@ ) {
                      $err = "Error in $f"."'s 'to_db' ".
                      "attribute :<p>\n $@\nPlease report this error.\n";
                      &CgiError($err);
                      &dbi_disconnect($db_insert, \$insert_th);
                      goto BAIL;
                   }
		}
		if (!$data) {
		   push( @insert_values, &dbi_get_null_assignment );
                } elsif ( $field{$f, 'type'} eq "char" 
	             || $field{$f, 'type'} eq "datetime" ) {
	           $data = &dbi_escape_characters($data);
	           $data = "'".$data."'";
	           push( @insert_values, $data );
	        } else {
	           push( @insert_values, $data );
		}
	    }
	}
    }

    #-------------------------------------------------------------------------
    # Construct the insert statement.  If this is a join form, break the TABLE
    # field into table names, and for each name find the fields that belong to
    # that table.  This will only work if:
    #   1. The serial key field is in the FDF before any foreign key,
# line 2050 WDB
    #   2. The table containing the serial key is first in the TABLE field,
    #   3. All foreign keys reference the serial key,
    #   4. Only one serial key exists in the FDF.
    #-------------------------------------------------------------------------
    if ( $form{'JOIN'} ) {
	@tables = split(',',$form{'TABLE'});

	# If there is a primary key value which is generated by the database,
	#  make sure the table containing that field is inserted into first so
	#  the new key value can be propogated to any foreign key fields.
	if ($post_insert eq "empty") {
	   @temp_tables = ();
	   foreach $t ( @tables ) {
	      if ($t eq $post_insert_table) {
	         unshift(@temp_tables, $t);
	      } else {
	         push(@temp_tables, $t);
	      }
	   }
	   @tables = ();
	   foreach $t (@temp_tables) {
	      push(@tables, $t);
	   }
	}

	# Construct an insert statement for each table in the JOIN.
	foreach $t ( @tables ) {
	    @insert_join_fields = ();
	    @insert_join_values = ();
	    $counter = 0;

	    foreach $f ( @insert_fields ) {

	        ($table_name, $field_name) = split( /\./, $f);

                # Eliminate leading and trailing blanks from the table name.
		$t =~ s/^\s*//;
		$t =~ s/\s*$//;

		# If foreign keys need to be filled before inserting, do that
		#  here.
		if ($table_name eq $t) {
		    push( @insert_join_fields, $field_name );

		    # Here's where we use the name of the foreign key fields
		    #  we saved earlier.
                    if ( defined $foreign_keys{$f}
		         && $post_insert eq "filled" ) {

# line 2100 WDB
		       push( @insert_join_values, $primary_key );
		    } else {
		       push( @insert_join_values, $insert_values[$counter] );
		    }
		}
		$counter++;
	    }

            $insert_into = 
	       "insert into $t (".join(',', @insert_join_fields).")";
            $value_list = " values (".join(',', @insert_join_values).")";

#$debug = 1;
            print "**** QUERY ****<br>" if $debug;
            print "$insert_into <br>" if $debug;
            print "$value_list <br>" if $debug;
#$debug = 0;

            $result = "";
            $result = 
	       &dbi_dosql($db_insert, \$insert_th, "$insert_into $value_list");

            if ($result) {
               print $result;
               &dbi_disconnect($db_insert, \$insert_th);
               goto BAIL;
            }

	    # If needed, get the newly generated serial key value.
	    if ($post_insert eq "empty") {
	       $primary_key = &dbi_get_generated_key($insert_th, $db_insert);
	       $val = $primary_key;
               if (defined $field{$f,'to_db'}) {
                  $primary_key = eval $field{$f,'to_db'};
                  if ( $@ ) {
                     $err = "Error in $f"."'s 'to_db' attribute :".
                            "<p>\n $@\nPlease report this error.\n";
                     &CgiError($err);
                     &dbi_disconnect($db_insert, \$insert_th);
                     goto BAIL;
                  }
               } 
	       print "Got serial key value: $primary_key<BR><BR>\n" if $debug;
	       $post_insert = "filled";
	    }

	}

    } else {
# line 2150 WDB
        $insert_into = 
	  "insert into $form{'TABLE'} (".join(',', @insert_fields).")";
        $value_list = " values (".join(',', @insert_values).")";

#$debug = 1;
        print "**** QUERY ****<br>" if $debug;
        print "$insert_into <br>" if $debug;
        print "$value_list <br>" if $debug;
#$debug = 0;

        $result = "";
        $result = 
	   &dbi_dosql($db_insert, \$insert_th, "$insert_into $value_list");

        if ($result) {
           print $result;
           &dbi_disconnect($db_insert, \$insert_th);
           goto BAIL;
        }

        # If needed, get the newly generated serial key value.
        if ($post_insert eq "empty") {
           $primary_key = &dbi_get_generated_key($insert_th, $db_insert);
           $val = $primary_key;
           print "Got serial key value: $primary_key<BR><BR>\n" if $debug;
           $post_insert = "filled";
        }

    }

    #-------------------------------
    # Disconnect from the database.
    #-------------------------------
    &dbi_disconnect($db_insert, \$insert_th);

    print "<br>The data has been inserted.<br>";

    if ( defined $primary && $debug ) {
	print"The primary key value is: $primary_key.<br>";
    }

    &html_tail unless defined $form{'NO_TAIL'};
    print "</BODY></HTML>\n";
}
#
#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#.PURPOSE	Performs a database update.
#------------------------------------------------------------------------------
sub DoUpdate 
# line 2200 WDB
{
    my (@update, @update_list, $f, $t, $keyval, $update, $die, $die2);
    my (@update_values, @update_fields, @tables, $table_name, $field_name);
    my (@join_fields, $err, @null_fields, $where, $old_data, $val, $ret);
    my ($data, $update_temp, $counter, $update_list, %serial_fields);
    my ($rowcount, $check_query, @update_titles);

    #--------------------------
    # Print HTML Header stuff.
    #--------------------------
    &write_header;
    &write_menu;

    #----------------------------------------------------------
    # Get qualification either from the form or from the path
    #----------------------------------------------------------
    &ReadParse(\%in, \@in, \$in);

    #------------------------------------------------------------------------
    # First make sure all required fields are filled.
    # If this is a join form, make sure all field names are fully qualified.
    # I don't care if the database will support non-qualified fields, because
    # the qualified field names are the only way I know which table to update
    # the data in.
    #------------------------------------------------------------------------
    $die = $die2 = "";

    foreach $f ( @fields ) {
        next if defined $field{$f,'computed'};

        if ( defined $field{$f, 'non_null'} && ( !defined $in{$f} ||
	     ( defined $in{$f} && $in{$f} eq "" )) ) {
            if ( !defined $die ) {
		$die = "TRUE";	
	    }
	    push( @null_fields, "$f" );
        }

        # This is a kludge to ignore the 'MORE' field.
	if ( $form{'JOIN'} && defined $field{$f, 'column'} ) {
	    ($table_name, $field_name) = split( /\./, $field{$f, 'column'});
	    if ( !defined $field_name ) {
		$die2 = "TRUE";
		push( @join_fields, "$f" );
	    }
	}
    }

    #--------------------------------------------------
# line 2250 WDB
    # If a required field is empty, or a field isn't qualified, say so.
    #--------------------------------------------------
    if ( $die ) {
	$err = "Required field(s): ".join(',',@null_fields)." cannot be empty.";
        print "$err<br>";
	return;
    }
    if ($die2) {
	$err = "field(s): ".join(',',@join_fields)." must be fully qualified";
        print "On an update, $err in the FDF.<br>";
	return;
    }

    # Get the original values for the update 'where' clause.  Do it here so 
    #  user's Perl code can refer to $old_values{'name'} if they need to.
    %old_values = ();
    foreach $f ( @fields ) {
	next if defined $field{$f, 'computed'};
	next if defined $field{$f, 'hidden'};

	$old_data = "";
	$old_data = $in{"old_value_".$f};
	
	if ( ($field{$f, 'type'} eq "char" || $field{$f, 'type'} eq "datetime")
	      && $old_data ne "NULL" ) {
	    $old_data = &dbi_escape_characters($old_data);
	    $old_data = "'".$old_data."'";
	}
	$old_values{$f} = "$old_data";
    }

    #--------------------------------------------------
    # Compose update fields and set list for the SQL statement.
    #--------------------------------------------------
    %serial_fields = ();
    foreach $f ( @fields ) {

	$val = "";
	$val = $in{$f} if defined $in{$f};

	if ( defined $field{$f,'computed'} ) { 
	   print "Evaluating computed field : $f = $in{$f}<BR>\n" if $debug;
	   print "- Compute : $field{$f,'to_db'}<BR>\n" if $debug;
	   $ret = eval $field{$f,'to_db'};
           if ( $@ ) {
              $err = "Error in $f"."'s 'to_db' attribute :".
                     "<p>\n $@\nPlease report this error.\n";
              &CgiError($err);
              &dbi_disconnect($db_update, \$update_th);
              goto BAIL;
           }
# line 2300 WDB
	   print " = $ret<BR>\n" if $debug;
	   next;	# Don't update computed fields.
	}

        #---------------------------------------------------------
	# NULL 'char' fields need to be compared using 'ne ""'.
	# NULL numerical fields need to be compared to 0.
        #---------------------------------------------------------
	unless ( defined $field{$f,'computed'} ) {
           if ( defined $in{$f} ) {
	      push( @update_fields, ( $field{$f,'column'} ) 
		    ?  $field{$f,'column'} : $f ); 

              $data = $in{$f};
              if (defined $field{$f,'to_db'}) {
                 $data = eval $field{$f,'to_db'};
                 if ( $@ ) {
                    $err = "Error in $f"."'s 'to_db' attribute :".
                           "<p>\n $@\nPlease report this error.\n";
                    &CgiError($err);
                    goto BAIL;
                 }
	      }
	      if ($field{$f, 'type'} eq "char" && 
	          ($data ne "NULL" && $data ne '') ) {
	         $data = &dbi_escape_characters($data);
	         $data = "'".$data."'";
	         push( @update_values, $data );
	      } elsif ($field{$f, 'type'} eq "char" && 
	              ($data eq "NULL" || $data eq '') ) {
	         push( @update_values, &dbi_get_null_assignment );
	      } elsif ($field{$f, 'type'} ne "char" && $data eq "") {
	         push( @update_values, &dbi_get_null_assignment );
	      } else {
	         push( @update_values, $data );
	      }

	   } else {
	      push( @update_fields, ( $field{$f,'column'} )
	            ?  $field{$f,'column'} : $f );
	      push( @update_values, &dbi_get_null_assignment );
	   }
	   # Keep track of serial fields.  Some databases won't allow you to
	   #  update a serial field.
	   if (defined $field{$f,'serial_field'}) {
	      $serial_fields{$field{$f,'column'}} = $f;
	   }
	}
    }
# line 2350 WDB

    if ( $form{'JOIN'} ) {
	@tables = split(',',$form{'TABLE'});

	foreach $t ( @tables ) {

            # Eliminate leading and trailing blanks from the table name.
            $t =~ s/^\s*//;
            $t =~ s/\s*$//;

	    @update = ();
	    @update_list = ();
	    $counter = 0;

	    $update = "update $t set ";

	    foreach $f ( @update_fields ) {
	        ($table_name, $field_name) = split( /\./, $f);

	        if (defined $serial_fields{$f}
		    && !&dbi_can_update_serial_field ) {
	           $counter++;
		   next;
		}
		if ($table_name eq $t) {
		    $update_temp = $field_name."=";
		    if ( defined $field{$f, 'type'} &&
		         $field{$f, 'type'} eq "char" ) {
			$data = "'".$update_values[$counter]."'";
			$update_temp = $update_temp.$data;
                    } else {
			$update_temp = $update_temp.$update_values[$counter];
		    }
	            push (@update_list, $update_temp);
		}
		$counter++;
	    }
	    $update_list = join(",\n", @update_list);

	    # This section creates the 'where' clause.  When a record is
	    #  is updated, all of the fields must be as they were when the
	    #  record was selected from the database for updating.  If anything
	    #  has changed, that means someone else updated or deleted the
	    #  record while you were messing around with it. :)  If changes
	    #  are present, the update doesn't happen.
	    @update_list = ();
            $where = "";
	    @update_titles = ();
	    foreach $f (@fields) {
# line 2400 WDB
	       next if defined $field{$f, 'computed'};
	       next if defined $field{$f, 'no_where'};
	       next if defined $field{$f, 'hidden'};

               ($table_name, $field_name) = split( /\./, $field{$f,'column'});

               if ($table_name eq $t) {
	       
		  push(@update_titles, $field{$f, 'column'});
		  if ($old_values{$f} eq "NULL") {
                     $update_temp 
		        = $field{$f, 'column'}.&dbi_get_null_comparator;
		  } else {
                     $update_temp = $field{$f, 'column'}."=";
		  }
                  if ( $field{$f, 'type'} eq "char" ) {
                     $data = "$old_values{$f}";
                     $update_temp = $update_temp.$data;
                  } else {
                     $update_temp = $update_temp.$old_values{$f};
                  }
                  push (@update_list, "$update_temp");
               }
	    }
	    $where = " where ".join(" and\n", @update_list);

#$debug = 1;
            print "**** QUERY ****<br>" if $debug;
            print "$update <br>" if $debug;
	    print "$update_list <br>" if $debug;
            print "$where<br>" if $debug;
#$debug = 0;

            #-------------------------------
            # Open the database.
            #-------------------------------
            $result = "";
            $result = &dbi_connect( $GLOBAL_USER, $GLOBAL_PSWD, $GLOBAL_SERVER, 
	                            $form{'DATABASE'}, \$db_update );

            if ($result) {
               print $result;
               &dbi_disconnect($db_update, \$update_th);
               goto BAIL;
            }

	    if ($form{'ALLOW_UPDATE'} ne "nocheck") {
               $rowcount = 0;
	       $check_query = "select count(*) from $t $where";
#$debug = 1;
               print "$check_query<br>\n" if $debug;
# line 2450 WDB
#$debug = 0;
	       $rowcount = 
	          &dbi_countrows($db_update, $check_query, @update_titles);

               if ($rowcount != 1) {
	           print "<B>ERROR:</B><BR><BR>\n\n".
		         "The data for this record has changed in the ".
		         "database since you selected it for an update!  This ".
		         "may have been caused by another database user ".
			 "altering the same information you are trying to ".
			 "change.<BR><BR>\n";
		   print "WDB cannot ".
		         "overwrite the changes.  You need to:";
		   print "<UL><LI>Use your browser's <B>Back</B> ".
		         "button to go back to the Update form.".
		         "<LI><B>RELOAD</B> the page to get the current data.".
		         "<LI>Make ".
		         "your changes and try to update the record again.".
		         "<LI>If this problem persists, please contact your ".
		         "database administrator.".
		         "</UL>\n";
                   &dbi_disconnect($db_update, \$update_th);
                   &html_tail unless defined $form{'NO_TAIL'};
                   print "</BODY></HTML>\n";
                   goto BAIL;
	       }
	    }

            $result = "";
	    $result = &dbi_dosql($db_update, \$update_th, 
	                         "$update $update_list $where");

            if ($result) {
               print $result;
               &dbi_disconnect($db_update, \$update_th);
               goto BAIL;
            }

            #-------------------------------
            # Disconnect from the database.
            #-------------------------------
            &dbi_disconnect($db_update, \$update_th);

	}
    } else {
	$update = "update $form{'TABLE'} set ";

	$counter = 0;
	for $f (@update_fields) {
	    if (defined $serial_fields{$f}
	        && !&dbi_can_update_serial_field ) {
# line 2500 WDB
	       $counter++;
	       next;
	    }

            $update_temp = $update_fields[$counter].
	                   "=".$update_values[$counter];
	    push (@update_list, $update_temp);
	    $update_temp = "";

            $counter++;
	}
	$update_list = join(",\n", @update_list);

	# This section creates the 'where' clause.  When a record is
	#  is updated, all of the fields must be as they were when the
	#  record was selected from the database for updating.  If anything
	#  has changed, that means someone else updated or deleted the
	#  record while you were messing around with it. :)  If changes
	#  are present, the update doesn't happen.
        $counter = 0;
	@update_list = ();
        foreach $f (@fields) {
	   next if $field{$f, 'computed'};
	   next if defined $field{$f, 'no_where'};
	   next if defined $field{$f, 'hidden'};

           if ($old_values{$f} eq "NULL") {
              $update_temp = $field{$f, 'column'}.&dbi_get_null_comparator;
           } else {
                $update_temp = $field{$f, 'column'}."=";
           }
           if ( $field{$f, 'type'} eq "char" ||
	        $field{$f, 'type'} eq "datetime" ) {
              $data = "$old_values{$f}";
              $update_temp = $update_temp.$data;
           } else {
              $update_temp = $update_temp.$old_values{$f};
           }
           push (@update_list, $update_temp) 
	      unless defined $field{$f, 'no_where'};
           $counter++;
        }
        $where = " where ".join(" and\n", @update_list);
#$debug = 1;
        print "**** QUERY ****<br>" if $debug;
        print "$update <br>" if $debug;
        print "$update_list <br>" if $debug;
	print "$where<br>" if $debug;
#$debug = 0;

        #-------------------------------
# line 2550 WDB
        # Open the database.
        #-------------------------------
        $result = "";
        $result = &dbi_connect( $GLOBAL_USER, $GLOBAL_PSWD, $GLOBAL_SERVER, 
	                        $form{'DATABASE'}, \$db_update );

        if ($result) {
           print $result;
           &dbi_disconnect($db_update, \$update_th);
           goto BAIL;
        }

        $rowcount = 0;
        $check_query = "select count(*) from $form{'TABLE'} $where";
#$debug = 1;
        print "$check_query<br>\n" if $debug;
#$debug = 0;
	if ($form{'ALLOW_UPDATE'} ne "nocheck") {
           $rowcount =
              &dbi_countrows($db_update, $check_query, @update_titles);

           if ($rowcount != 1) {
               print "<B>ERROR:</B><BR><BR>\n\n".
                     "The data for this record has changed in the database ".
                     "since you selected it for an update!  This may have ".
                     "been caused by another database user altering the same ".
                     "information you are trying to change.<BR><BR>\n";
               print "WDB cannot ".
                     "overwrite the changes.  You need to:";
               print "<UL><LI>Use your browser's <B>Back</B> ".
                     "button to go back to the Update form.".
                     "<LI><B>RELOAD</B> the page to get the current data.".
                     "<LI>Make ".
                     "your changes and try to update the record again.".
                     "<LI>If this problem persists, please contact your ".
                     "database administrator.".
                     "</UL>\n";
               &dbi_disconnect($db_update, \$update_th);
               &html_tail unless defined $form{'NO_TAIL'};
               print "</BODY></HTML>\n";
               goto BAIL;
           }
	}

        $result = "";
        $result = 
	   &dbi_dosql($db_update, \$update_th, "$update $update_list $where");

        if ($result) {
           print $result;
           &dbi_disconnect($db_update, \$update_th);
# line 2600 WDB
           goto BAIL;
        }

        #-------------------------------
        # Disconnect from the database.
        #-------------------------------
        &dbi_disconnect($db_update, \$update_th);

    }

    print "<br>The data has been updated.<br>";

    &html_tail unless defined $form{'NO_TAIL'};
    print "</BODY></HTML>\n";

}
#
#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#.PURPOSE	Performs a database update.
#------------------------------------------------------------------------------
sub DoDelete 
{
    my ($delete_temp, $where, $data, $delete, $foreign_key, @delete_list);
    my ($username, $password, $dbpassword, @tables, $f, $t, $ret, $err);
    my (@delete, $counter, $table_name, $field_name, $old_data);
    my ($rowcount, $check_query, @delete_titles);

    #--------------------------
    # Print HTML Header stuff.
    #--------------------------
    &write_header;
    &write_menu;

    #----------------------------------------------------------
    # Get qualification either from the form or from the path
    #----------------------------------------------------------
    &ReadParse(\%in, \@in, \$in);

    #-------------------------------
    # Get values of old fields.  If users change any of the fields in the form,
    #  too bad, we don't use them.
    #-------------------------------
    foreach $f ( @fields ) {
        next if defined $field{$f, 'computed'};

	# Get the original values for the update 'where' clause.
	$old_data = "";
	$old_data = $in{"old_value_".$f};
	
# line 2650 WDB
	if ( ($field{$f, 'type'} eq "char" || $field{$f, 'type'} eq "datetime")
	      && $old_data ne "NULL" ) {
	    $old_data = &dbi_escape_characters($old_data);
	    $old_data = "'".$old_data."'";
	}
	$old_values{$f} = $old_data;

    }

    #----------
    # Paranoid people (like me) don't let just anyone whack data.  You can
    #  use a username/password protection to protect deletes.  See the
    #  dbi_getpassword() function in whichever dbi file you are using to see
    #  how this needs to be done, or feel free to write your own.
    #----------
    if ( defined $form{'PASSWORD_DELETE'} ) {
        $username = $password = "";
        if ( defined $in{"username"} && $in{"username"} ne "") {
	    $username = $in{"username"};
        }
        if ( defined $in{"password"} && $in{"password"} ne "") {
	    $password = $in{"password"};
        }
        if ( !defined $username || !defined $password ) {
	    print "<B>Invalid username or password.</B><BR>";
	    return;
        }
    }

    #-------------------------------
    # Open the database.
    #-------------------------------
    $result = "";
    $result = 
       &dbi_connect( $GLOBAL_USER, $GLOBAL_PSWD, $GLOBAL_SERVER, 
                     $form{'DATABASE'}, \$db_delete );

    if ($result) {
       print $result;
       &dbi_disconnect($db_delete, \$delete_th);
       goto BAIL;
    }

    if ( defined $form{'PASSWORD_DELETE'} ) {
        $dbpassword = &dbi_getpassword($db_delete, $username);

        if ( $password ne $dbpassword ) {
            print "<B>Invalid username or password.</B><BR>";
            return;
        }
# line 2700 WDB
    }

    #-----------
    # If there is a join condition, delete records in all affected tables.  If
    #  you only want to delete from one table, CREATE ANOTHER FDF and use it!
    #-----------
    if ( $form{'JOIN'} ) {
	@tables = split(',',$form{'TABLE'});

	foreach $t ( @tables ) {

            $delete = "delete from $t ";

	    @delete = ();
	    @delete_list = ();
	    $counter = 0;
            $where = "";
            foreach $f (@fields) {

	       if ( defined $field{$f,'computed'} ) { 
	          print "Evaluating computed field : $f = $in{$f}<BR>\n" if $debug;
	          print "- Compute : $field{$f,'to_db'}<BR>\n" if $debug;
	          $ret = eval $field{$f,'to_db'};
                  if ( $@ ) {
                     $err = "Error in $f"."'s 'to_db' attribute :".
                            "<p>\n $@\nPlease report this error.\n";
                     &CgiError($err);
                     &dbi_disconnect($db_delete, \$delete_th);
                     goto BAIL;
                  }
	          print " = $ret<BR>\n" if $debug;
	          next;	# Don't delete computed fields.
	       }

               ($table_name, $field_name) = split( /\./, $field{$f,'column'});

               # Eliminate leading and trailing blanks from the table name.
               $t =~ s/^\s*//;
               $t =~ s/\s*$//;

               if ($table_name eq $t) {
		  if ($old_values{$f} eq "NULL") {
                     $delete_temp 
		        = $field{$f, 'column'}.&dbi_get_null_comparator;
		  } else {
                     $delete_temp = $field{$f, 'column'}."=";
		  }
                  if ( $field{$f, 'type'} eq "char" ) {
                     $data = "$old_values{$f}";
# line 2750 WDB
                     $delete_temp = $delete_temp.$data;
                  } else {
                     $delete_temp = $delete_temp.$old_values{$f};
                  }
                  push (@delete_list, $delete_temp) 
		     unless defined $field{$f, 'no_where'};
               }
               $counter++;
            }
            $where = " where ".join(" and\n", @delete_list);
#$debug = 1;
            print "**** QUERY ****<br>" if $debug;
            print "$delete <br>" if $debug;
            print "$where<br>" if $debug;
#$debug = 0;

	    if ($form{'ALLOW_DELETE'} ne "nocheck") {
               $rowcount = 0;
               $check_query = "select count(*) from $t $where";
#$debug = 1;
               print "$check_query<br>\n" if $debug;
#$debug = 0;
               $rowcount =
                  &dbi_countrows($db_delete, $check_query, @delete_titles);

               if ($rowcount != 1) {
                   print "<B>ERROR:</B><BR><BR>\n\n".
                         "The data for this record has changed in the ".
                         "database since you selected it for deletion!  This ".
                         "may have been caused by another database user ".
                         "altering or deleting the same ".
                         "information you are trying to delete.<BR><BR>\n";
                   print "WDB cannot ".
                         "overwrite the changes.  You need to:";
                   print "<UL><LI>Use your browser's <B>Back</B> ".
                         "button to go back to the Delete form.".
                         "<LI><B>RELOAD</B> the page to get the current data.";
                   print "<LI>If no data is found, then someone else has ".
		         "already deleted the data.";
                   print "<LI>If the data reloads, try to delete the record ".
		         "again.".
                         "<LI>If this problem persists, please contact your ".
                         "database administrator.".
                         "</UL>\n";
                   &dbi_disconnect($db_delete, \$delete_th);
                   &html_tail unless defined $form{'NO_TAIL'};
                   print "</BODY></HTML>\n";
                   goto BAIL;
               }
	    }

            $result = "";
# line 2800 WDB
            $result = &dbi_dosql($db_delete, \$delete_th, "$delete $where");

            if ($result) {
               print $result;
               &dbi_disconnect($db_delete, \$delete_th);
               goto BAIL;
            }
	}
    } else {

	$delete = "delete from $form{'TABLE'} ";

        $where = "";
        $counter = 0;
	@delete_list = ();
        foreach $f (@fields) {
	
          next if defined $field{$f, 'computed'};

          if ($old_values{$f} eq "NULL") {
             $delete_temp = $field{$f, 'column'}.&dbi_get_null_comparator;
          } else {
             $delete_temp = $field{$f, 'column'}."=";
          }

          if ( $field{$f, 'type'} eq "char" ) {
             $data = "$old_values{$f}";
             $delete_temp = $delete_temp.$data;
          } else {
             $delete_temp = $delete_temp.$old_values{$f};
          }
          push (@delete_list, $delete_temp)
	      unless defined $field{$f, 'no_where'};
	  $counter++;
        }
        $where = " where ".join(" and\n", @delete_list);

#$debug = 1;
        print "**** QUERY ****<br>" if $debug;
        print "$delete <br>" if $debug;
	print "$where<br>" if $debug;
#$debug = 0;

	if ($form{'ALLOW_DELETE'} ne "nocheck") {
           $rowcount = 0;
           $check_query = "select count(*) from $form{'TABLE'} $where";
#$debug = 1;
           print "$check_query<br>\n" if $debug;
#$debug = 0;
           $rowcount =
# line 2850 WDB
              &dbi_countrows($db_delete, $check_query, @delete_titles);

           if ($rowcount != 1) {
               print "<B>ERROR:</B><BR><BR>\n\n".
                     "The data for this record has changed in the database ".
                     "since you selected it for deletion!  This may have ".
                     "been caused by another database user altering or ".
                     "deleting the same ".
                     "information you are trying to delete.<BR><BR>\n";
               print "WDB cannot ".
                     "overwrite the changes.  You need to:";
               print "<UL><LI>Use your browser's <B>Back</B> ".
                     "button to go back to the Delete form.".
                     "<LI><B>RELOAD</B> the page to get the current data.";
               print "<LI>If no data is found, then someone else has already ".
                     "deleted the data.";
               print "<LI>If the data reloads, try to delete the record again.".
                     "<LI>If this problem persists, please contact your ".
                     "database administrator.".
                     "</UL>\n";
               &dbi_disconnect($db_delete, \$delete_th);
               &html_tail unless defined $form{'NO_TAIL'};
               print "</BODY></HTML>\n";
               goto BAIL;
           }
        }

        $result = "";
        $result = &dbi_dosql($db_delete, \$delete_th, "$delete $where");

        if ($result) {
           print $result;
           &dbi_disconnect($db_delete, \$delete_th);
           goto BAIL;
        }
    }

    #-------------------------------
    # Disconnect from the database.
    #-------------------------------
    &dbi_disconnect($db_delete, \$delete_th);

    print "<br>The data has been deleted.<br>";

    &html_tail unless defined $form{'NO_TAIL'};
    print "</BODY></HTML>\n";
}
#
#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#.PURPOSE       Print variables in 'full-screen' format.  
# line 2900 WDB
#------------------------------------------------------------------------------
sub PrintVars (\%)
{
    my (%val) = @_;
    my ($rectop, $max, $max2, $f, $column, $value, $sub_query_header, $table);
    my (@my_fields, %my_field, %my_form, @my_keys, @my_select_titles);
    my ($fdf_file, $tmp_query, $sub_query, $sub_qh, $my_f, %my_row);
    my ($my_max, $my_max2);

    #-------------
    # Print header
    #-------------
    if ( $form{'RECTOP'} ) {
	$rectop = eval $form{'RECTOP'};
	if ( $@ ) {
	    &CgiError("Error in the 'RECTOP' attribute :<p>\n $@\n");
            &dbi_disconnect($db_query, \$main_th);
	    goto BAIL;
	}
	print "<h2>$rectop</h2>\n";
    }
    if ( $form{'TABLE_RESULTS'} ) {
        print "\n<TABLE";
	print " ".$form{'TABLE_RESULTS'} if $form{'TABLE_RESULTS'} ne "TRUE";
	print ">";
    } else {
        #----------------------------
        # Find max length of titles.
        #----------------------------
        $max = 0;
        $max2 = 0;
        foreach $f ( @fields ) {
	    next if defined $field{$f,'no_label'};
	    next if defined $field{$f,'no_dots'};
	    next if defined $field{$f,'ignore_select'};
	    next if defined $field{$f,'hidden'};
            $max = length($field{$f,'label'}) 
	      if length($field{$f,'label'}) > $max;
            if ( defined $field{$f,'tablen'} ) {
               $max2 = $field{$f,'tablen'} if $field{$f,'tablen'} > $max2;
            } else {
               $max2 = $field{$f,'length'} if $field{$f,'length'} > $max2;
            }
        }
        $max2 = $max2 + 1;

        if (defined $form{'PRE_RESULTS'}) {
            print "<PRE>";
        } else {
# line 2950 WDB
            print "\n<UL>";
	}
    }

    #-----------------------
    # Now print each field.
    #-----------------------
    foreach $f ( @fields ) {
        next if !defined $f || $f eq "";
	next if defined $field{$f,'hidden'};	# Ignore hidden fields
	next if defined $field{$f,'no_full'};	# Ignore no_full fields
	next if defined $field{$f,'ignore_select'};

        PrintVarsRow($max, $max2, $f, \%val, \%form, \%field)
          if !defined $field{$f,'sub_query_pre'};

        @my_fields = @my_keys = @my_select_titles = ();
	%my_field = %my_form = %my_row = ();
	$fdf_file = $tmp_query = $sub_query = $sub_qh = $my_f = 0;
	$my_max = $my_max2 = 0;

	# Perform a sub-query and display the results.
	if ( defined $field{$f,'sub_query'} ) {
	   ($fdf_file, $tmp_query) = split('\|', $field{$f,'sub_query'});
	   $sub_query = eval $tmp_query;
           if ( $@ ) {
               &CgiError("Error in the 'sub_query' attribute :<p>\n $@\n");
               &dbi_disconnect($db_query, \$main_th);
               goto BAIL;
           }

	   &ParseFile("$FORMDIR/$DBDIR/$fdf_file".".fdf", \@my_fields,
	              \%my_field, \@my_keys, \%my_form);

           $my_max = 0;
           $my_max2 = 0;
	   foreach $my_f (@my_fields) {
              next if defined $my_field{$my_f,'computed'};
              next if defined $my_field{$my_f,'ignore_select'};
	      push(@my_select_titles, $my_f);
              next if defined $my_field{$my_f,'hidden'};
              next if defined $my_field{$my_f,'no_label'};
              next if defined $my_field{$my_f,'no_dots'};

              $my_max = length($my_field{$my_f,'label'})
                if length($my_field{$my_f,'label'}) > $my_max;
              if ( defined $my_field{$my_f,'tablen'} ) {
                 $my_max2 = $my_field{$my_f,'tablen'}
                   if $my_field{$my_f,'tablen'} > $my_max2;
              } else {
                 $my_max2 = $my_field{$my_f,'length'}
                   if $my_field{$my_f,'length'} > $my_max2;
              }
           }
# line 3000 WDB
           $my_max2 = $my_max2 + 1;

	   $sub_query = "select ".join(',', @my_select_titles)." from ".
			"$my_form{'TABLE'} where $sub_query";

	   &dbi_dosql($db_query, \$sub_qh, $sub_query);
	   if ( (%my_row = &dbi_nextrow($sub_qh, @my_select_titles) ) ) {
	      if (defined $my_form{'SUB_QUERY_HEADER'}) {
	         $sub_query_header = eval $my_form{'SUB_QUERY_HEADER'};
                 if ( $@ ) {
                     &CgiError("Error in the 'SUB_QUERY_HEADER' attribute :".
			       "<p>\n $@\n");
                     &dbi_disconnect($db_query, \$main_th);
                     goto BAIL;
                 }
		 print $sub_query_header;
	      }
	      do {
	         foreach $my_f (@my_fields) {
                    next if !defined $my_f || $my_f eq "";
                    next if defined $my_field{$my_f,'hidden'};
                    next if defined $my_field{$my_f,'no_full'};
                    next if defined $my_field{$my_f,'ignore_select'};

	            PrintVarsRow($max, $max2, $my_f, \%my_row, 
				 \%my_form, \%my_field);
	         }
	      } while (%my_row = &dbi_nextrow($sub_qh, @my_select_titles));
	   }
	   $sub_qh->finish if defined $sub_qh;
	}

        PrintVarsRow($max, $max2, $f, \%val, \%form, \%field)
           if defined $field{$f,'sub_query_pre'};

    }
    if ( $form{'TABLE_RESULTS'} ) {
        print "\n</TABLE>\n";
    } else {
        if (defined $form{'PRE_RESULTS'}) {
            print "</PRE>\n";
        } else {
            print "\n</UL>\n";
	}
    }
    # This bar goes in between records.
    print "<HR>";
}
#
# line 3050 WDB
#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#.PURPOSE	Prints a row for full-screen output
#
#.REMARKS       I broke this out of PrintVars to handle sub_queries.
#------------------------------------------------------------------------------
sub PrintVarsRow 
{
   my($my_max, $my_max2, $f, $val_ref, $my_form, $my_field) = @_;
   my($pre_html, $dummy, $column, $value, %val, @sql_dos, $sql_check);
   my($sql_counter, $sql_query, $num_rows, $output, @sql_checks);

   # This is required to use $val{'whatever'} in FDF attributes.
   %val = %$val_ref;

   # This uses the name of the column, not the name of the field.
   #  If I change the FIELD name, database queries don't work, but the
   #  'column' field must be the name used in the database query.
   # Note: this means you MUST use $val{'t1.f1'} type of notation when
   #  using table joins.
   $column = $$my_field{$f, 'column'};

   # Put results in an HTML table.
   if ( $$my_form{'TABLE_RESULTS'} ) {
      print "\n<TR>"; 
      if (defined $$my_field{$f,'pre_html'}) {
         $pre_html =  eval $$my_field{$f,'pre_html'};
         if ( $@ ) {
            &CgiError("Error in the 'pre_html' attribute :<p>\n $@\n");
            &dbi_disconnect($db_query, \$main_th);
            goto BAIL;
         }
         print "$pre_html";
      }

      if (! defined $$my_field{$f,'no_label'}) {
          print "\n   <TH"; 
          if ( defined $$my_field{$f, 'header_format'} ) {
              print " $$my_field{$f, 'header_format'}"
          } elsif ( defined $$my_form{'HEADER_FORMAT'} ) {
              print " $$my_form{'HEADER_FORMAT'}"
          }
          print ">";
          print $$my_field{$f,'label'};
	  print "</TH>\n";
      }
      $value = 
         &Value( $f, \$val{($column ? $column : $f)}, \%val, $my_field );
      $value =~ s/^\s*//;
      $value =~ s/\s*$//;
# line 3100 WDB

      print "   <TD";
      if ( defined $$my_field{$f, 'cell_format'} ) {
         print " $$my_field{$f, 'cell_format'}"
      } elsif ( defined $$my_form{'CELL_FORMAT'} ) {
         print " $$my_form{'CELL_FORMAT'}"
      }
      if ($$my_field{$f, 'type'} eq "char" && ($value eq " " || $value eq "")) {
         print "><BR>";
      } elsif ( ($$my_field{$f, 'type'} eq "int"
                 || $$my_field{$f, 'type'} eq "float") 
		 && !defined $val{$column} ) {
         print "><BR>"; 
      } else {
         print ">$value";
      }

      print " $$my_field{$f,'unitlabel'}" 
        if ( $$my_field{$f,'unitlabel'} && $$my_field{$f} ne "" );
      print "</TD>\n";
      print "</TR>"; 

   } else {
      if (! defined $$my_form{'PRE_RESULTS'}) {
         print "\n<LI>" if !defined $$my_field{$f,'sameline_out'};
      } else {
         print "\n" if !defined $$my_field{$f,'sameline_out'};
      }

      if (defined $$my_field{$f,'pre_html'}) {
         $pre_html =  eval $$my_field{$f,'pre_html'};
         if ( $@ ) {
            &CgiError("Error in the 'pre_html' attribute :<p>\n $@\n");
            &dbi_disconnect($db_query, \$main_th);
            goto BAIL;
         }
         print $pre_html;
      }
       
      print " " x ($$my_field{$f,'pre_space'}) 
         if defined $$my_field{$f,'pre_space'};

      if (! defined $$my_field{$f,'no_label'}) {
         print "<TT>";
       	 print "<b>" if !defined $my_form->{'NO_BOLD_LABEL'};
         print $$my_field{$f,'label'};
         print "." x ($my_max - length($$my_field{$f,'label'}))
	    if ! defined $$my_field{$f,'no_dots'};
       	 print ":";
# line 3150 WDB
         print "</b>" if !defined $my_form->{'NO_BOLD_LABEL'};
         print "</TT> ";
      }
      $value = 
         &Value( $f, \$val{($column ? $column : $f)}, \%val, $my_field );
       
      $value = &cgi_encode($value) if defined $$my_field{$f,'encode'};
      print "$value";

      print " $$my_field{$f,'unitlabel'}" 
         if defined $$my_field{$f,'unitlabel'};

      print " " x ($$my_field{$f,'post_space'}) 
         if defined $$my_field{$f,'post_space'};

      if ( ! defined $$my_field{$f,'no_space'} ) {
         if ( defined $$my_field{$f,'url'} ) {
            print " " x ($my_max2 - $$my_field{$f,'length'});
         } else {
            print " " x ($my_max2 - length($value));
         }
       
      } elsif ( defined $$my_field{$f,'separator'} ) {
         print " $$my_field{$f,'separator'} " if $value ne "";
      } else {
         print " " unless defined $my_form->{'PRE_RESULTS'};
      }

      #-------------
      # Execute a sub query as a check before displaying something.
      #-------------
      if (defined $$my_field{$f,'sql_check'} 
          && defined $$my_field{$f,'sql_do'}) {

	  @sql_checks = split('\|', $$my_field{$f,'sql_check'});
	  @sql_dos = split('\|', $$my_field{$f,'sql_do'});

	  # Execute the sql statement
	  $sql_counter = 0;
	  foreach $sql_check (@sql_checks) {
	     print "$sql_check<BR>" if $debug;
	     $sql_query = eval "$sql_check";
	     if ( $@ ) {
	        &CgiError("Error in $f's 'sql_check' attribute :<p>\n ".
			  "$@\nPlease report this error.\n");
                &dbi_disconnect($db_query, \$main_th);
	        goto BAIL;
	     }
	     print "$sql_query<BR>" if $debug;
# line 3200 WDB
	     if ( ($num_rows = 
	           &dbi_countrows($db_query, $sql_query, @select_titles)) > 0) {

	        # If the results are positive, eval the next step
		print "$sql_dos[$sql_counter]<BR>" if $debug;
		$output = eval $sql_dos[$sql_counter];
	        if ( $@ ) {
	           &CgiError("Error in $f's 'sql_do' attribute :<p>\n ".
			     "$@\nPlease report this error.\n");
                   &dbi_disconnect($db_query, \$main_th);
	           goto BAIL;
	        }
		print "$output";
		#last;
	     }
	     $sql_counter++;
	  }
      }

      #--------------
      # Put HTML stuff after everything else.
      #--------------
      if (defined $$my_field{$f,'post_html'}) {
         $dummy = eval $$my_field{$f,'post_html'};
         if ( $@ ) {
            &CgiError("Error in $f 's 'post_html' attribute :".
                      "<p>\n $@\nPlease report this error.\n");
            &dbi_disconnect($db_query, \$main_th);
            goto BAIL;
         }
         print $dummy;
      }
      print "<BR>" if defined $$my_field{$f,'line_break'};
   }
}
#
#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#.PURPOSE	Print column headers for tabular output.
#------------------------------------------------------------------------------
sub PrintTabHead 
{
    my( $len, $f);
    my($dummy);

    if ( defined $form{'TABLE_RESULTS'} ) {
        print "\n<TABLE";
	print " ".$form{'TABLE_RESULTS'} if $form{'TABLE_RESULTS'} ne "TRUE";
	print ">\n<TR>";
    } else {
# line 3250 WDB
      print "<pre><b>\n";
    }

    foreach $f ( @fields ) {
	#------------
	# This is where the 'Update' and 'Delete' hyperlinks are printed, if
	#  the form allows it and the correct copy of WDB is being used.
	#------------
	if ( $f =~ /update/i ) {
            next if !defined $form{'ALLOW_UPDATE'};
	    next if !$ALLOW_MASTER && $ACTION eq "query";
	}
	if ( $f =~ /delete/i ) {
            next if !defined $form{'ALLOW_DELETE'};
	    next if !$ALLOW_MASTER && $ACTION eq "query";
	}
	$dummy = &FieldInTab( $f );
	if ( &FieldInTab( $f ) ) {
            if ( $form{'TABLE_RESULTS'} ) {
               print "\n   <TH";
               if ( defined $field{$f, 'header_format'} ) {
                  print " $field{$f, 'header_format'}"
               } elsif ( defined $form{'HEADER_FORMAT'} ) {
                  print " $form{'HEADER_FORMAT'}"
               }
	       print ">";
               print $field{$f,'label'};
	       print "</TH>";
            } else {
                if ( $output_mode eq 'tab' && $field{$f,'tablen'} ) {
                    $len = $field{$f,'tablen'};
                } else {
                    $len = $field{$f,'length'};
                }
                printf "%-$len.$len"."s ", $field{$f,'label'};
            }
	}
    }
    if ( $form{'TABLE_RESULTS'} ) {
        print "\n</TR>\n";
    } else {
        print "</b>\n";
    }
}
#
#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#.PURPOSE	Print a row in tabular output format.
#------------------------------------------------------------------------------
sub PrintTabVars (\%)
# line 3300 WDB
{
    my (%val) = @_;
    my (@my_fields, %my_field, %my_form, @my_keys, @my_select_titles);
    my ($fdf_file, $tmp_query, $sub_query, $sub_qh, $my_f, %my_row);
    my ($sub_query_header, $f);

    foreach $f ( @fields ) {

	if ( $f =~ /update/i ) {
            next if !defined $form{'ALLOW_UPDATE'};
	    next if !$ALLOW_MASTER && $ACTION eq "query";
	}
	if ( $f =~ /delete/i ) {
            next if !defined $form{'ALLOW_DELETE'};
	    next if !$ALLOW_MASTER && $ACTION eq "query";
	}
	next if !&FieldInTab( $f );

	&PrintTabVarsRow($f, \%val, \%form, \%field)
	   if !defined $field{$f,'sub_query_pre'};

        @my_fields = @my_keys = @my_select_titles = ();
	%my_field = %my_form = %my_row = ();
	$fdf_file = $tmp_query = $sub_query = $sub_qh = $my_f = 0;

        # Perform a sub-query and display the results.
        if ( defined $field{$f,'sub_query'} ) {
           ($fdf_file, $tmp_query) = split('\|', $field{$f,'sub_query'});
           $sub_query = eval $tmp_query;
           if ( $@ ) {
               &CgiError("Error in the 'sub_query' attribute :<p>\n $@\n");
               &dbi_disconnect($db_query, \$main_th);
               goto BAIL;
           }

           &ParseFile("$FORMDIR/$DBDIR/$fdf_file".".fdf", \@my_fields,
                      \%my_field, \@my_keys, \%my_form);

           foreach $my_f (@my_fields) {
              next if defined $my_field{$my_f,'computed'};
              next if defined $my_field{$my_f,'ignore_select'};
              push(@my_select_titles, $my_f);
           }

           $sub_query = "select ".join(',', @my_select_titles)." from ".
                        "$my_form{'TABLE'} where $sub_query";

           &dbi_dosql($db_query, \$sub_qh, $sub_query);
           if ( (%my_row = &dbi_nextrow($sub_qh, @my_select_titles) ) ) {
              if ( !defined $my_form{'TABLE_RESULTS'} ) {
                  print "\n";
              }
              if (defined $my_form{'SUB_QUERY_HEADER'}) {
# line 3350 WDB
                 $sub_query_header = eval $my_form{'SUB_QUERY_HEADER'};
                 if ( $@ ) {
                     &CgiError("Error in the 'SUB_QUERY_HEADER' attribute :".
                               "<p>\n $@\n");
                     &dbi_disconnect($db_query, \$main_th);
                     goto BAIL;
                 }
                 print $sub_query_header;
              }
              do {
                 if ( !defined $my_form{'TABLE_RESULTS'} ) {
                     print "\n";
                 }
                 foreach $my_f (@my_fields) {
                    next if !defined $my_f || $my_f eq "";
                    next if defined $my_field{$my_f,'hidden'};
                    next if defined $my_field{$my_f,'no_full'};
                    next if defined $my_field{$my_f,'ignore_select'};

                    PrintTabVarsRow($my_f, \%my_row, \%my_form, \%my_field);
                 }
              } while (%my_row = &dbi_nextrow($sub_qh, @my_select_titles));
           }
           $sub_qh->finish if defined $sub_qh;
        }

	&PrintTabVarsRow($f, \%val, \%form, \%field)
           if defined $field{$f,'sub_query_pre'};

    }
    if ( !defined $form{'TABLE_RESULTS'} ) {
        print "\n";
    }
}
#
#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#.PURPOSE	Print a row in tabular output format.
#
#.REMARKS       I broke this out of PrintTabVars to handle sub_queries.
#------------------------------------------------------------------------------
sub PrintTabVarsRow
{
   my($f, $val_ref, $my_form, $my_field) = @_;
   my(%val, $value, $column);

   # This is required to use $val{'whatever'} in FDF attributes.
   %val = %$val_ref;

   # This uses the name of the column, not the name of the field.
# line 3400 WDB
   #   If I change the FIELD name, database queries don't work, but the
   #   'column' field must be the name used in the database query.
   $column = $$my_field{$f, 'column'};
   $value = 
      &Value( $f, \$val{($column ? $column : $f)}, \%val, $my_field );

   if ( $$my_form{'TABLE_RESULTS'} ) {
      $value =~ s/^\s*//;
      $value =~ s/\s*$//;

      print "   <TD";
      if ( defined $$my_field{$f, 'cell_format'} ) {
         print " $$my_field{$f, 'cell_format'}"
      } elsif ( defined $$my_form{'CELL_FORMAT'} ) {
         print " $$my_form{'CELL_FORMAT'}"
      }
      if ( defined $$my_field{$f, 'computed'}) {
         print "> $value </TD>\n";
      } elsif (!defined $$my_field{$f, 'type'} || !defined $value) {
         print "> <br> </TD>\n";
      } elsif ($$my_field{$f, 'type'} eq "char"
               && ($value eq " " || $value eq "") ) {
         print "> <br> </TD>\n";
      } elsif ( ($$my_field{$f, 'type'} eq "int"
                  || $$my_field{$f, 'type'} eq "float") 
	          && !defined $val{$column} ) {
         print "> <br> </TD>\n"; 
      } else {
         print "> $value </TD>\n";
      }
   } else {
      print "$value ";
   }
}
#
#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#.PURPOSE	Decides if a field should be included in tabular output.
#
#.RETURNS	0 or 1 for yes or no.  
#------------------------------------------------------------------------------
sub FieldInTab (\$)
{
    my ($f) = @_;	# Field name
    my ($ec, @ec);

    return 1 if defined $field{$f,'forcetab'};
    return 1 if defined $in{"tab_$f"};
    if (defined $in{'extra_columns'}) {
       return 1 if $in{'extra_columns'} =~ /(\\0|^)$f(\\0|$)/;
# line 3450 WDB
       $ec = "";
       @ec = ();
       @ec = split('\0', $in{'extra_columns'});
       foreach $ec (@ec) {
          return 1 if $ec eq $f;
       }
    }
    if ( ($ACTION eq "update_query" || $ACTION eq "delete_query")
          && defined $in{$f} ) { return 1; }
    if ( $KEY_ARGS && $ACTION eq "page" ) { return 0; }
    if ( $KEY_ARGS || $ACTION eq "default" || $ENV{'REQUEST_METHOD'} eq 'GET'){
	return 0 if defined $field{$f,'hidden'};
	return 1 if ! defined $field{$f,'no_query'} 
		 && ! defined $field{$f,'no_tab'};
    }
    return 0;
}
#
#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#.PURPOSE	Compute output value of field (eval enums, pictures etc. ..)
#
#.RETURNS	New value;
#------------------------------------------------------------------------------
sub Value
{
    my( $f, $old_val, $old_val_hash, $my_field ) = @_;
    my ($url, $value, $val, $length, %val);
    $value = "";

    if ( defined $$old_val ) {
       $val = $$old_val;
    } else {
       $val = "";
    }
    if ( defined %$old_val_hash ) {
       %val = %$old_val_hash;
    } else {
       %val = ();
    }

    if ( $val eq "NULL" 
	|| (defined $$my_field{$f,'type'}
	    && defined $SPEC_NULL{$$my_field{$f,'type'}} 
	    && $SPEC_NULL{$$my_field{$f,'type'}} eq $val) ) {

	$value = $SPEC_NULL{$$my_field{$f,'type'}};
	if ( $output_mode eq 'tab' && defined $$my_field{$f,'tablen'} ) {
	    $value = sprintf("%-$$my_field{$f,'tablen'}.$$my_field{$f,'tablen'}s", $value);
	} else {
# line 3500 WDB
	    $value = sprintf("%-$$my_field{$f,'length'}s", $value);
	}
	return $value;
    } 

    if ( defined $$my_field{$f,'from_db'} ) {
	$value = eval $$my_field{$f,'from_db'};
	if ( $@ ) {
	    &CgiError("Error in $f 's 'from_db' attribute :".
		      "<p>\n $@\nPlease report this error.\n");
            &dbi_disconnect($db_query, \$main_th);
	    goto BAIL;
	}
    }

    #------------
    # This is a kludge to handle images in a database.  It would be easy to
    #  just select the data from the database and spit it out at the browser
    #  to display, except the browser has to be told the data is a binary
    #  image and the only way to tell it that is with the <IMG> tag.
    # So, image_src can be the URL to a picture, but it can also be the URL
    #  to an external cgi program that returns an image as the output of the
    #  program.  Either can be used in an <IMG> tag.
    # For example, image_src could be something like this:
    #  image_src = "SRC=http://somewhere/cgi-bin/get_image.cgi/$val{'key'}"
    #  The get_image.cgi program would select the image(s) based on the key
    #  value, and spit out a Content-type header and the binary data for the
    #  image.
    #------------
    if (defined $$my_field{$f,'image_src'}) {
       $value = eval $$my_field{$f,'image_src'};
       if ( $@ ) {
          &CgiError("Error in $f's 'image_src' attribute :".
	            "<p>\n $@\nPlease report this error.\n");
          &dbi_disconnect($db_query, \$main_th);
          goto BAIL;
       }
       $value = "<IMG $value>";
    } else {

       #-------------------------------
       # Adjust output length of value
       #-------------------------------
       if ($value eq "") { $value = $val; }
       $value =~ s/^\s*// if !defined $$my_field{$f,'from_db'};
       $value =~ s/\s*$// if !defined $$my_field{$f,'from_db'};
       if ( $output_mode eq 'tab' && defined $$my_field{$f,'tablen'} ) {
   	 $value = sprintf("%-$$my_field{$f,'tablen'}.$$my_field{$f,'tablen'}s", $value);
       } else {
# line 3550 WDB
   	 $value = sprintf("%-$$my_field{$f,'length'}s", $value);
       }
    }

    # Don't want URLs printed on the update or delete forms.
    if ( $$my_field{$f,'url'} && ($ACTION =~ /.*query$/ || $ACTION eq "page") ) {
	$url = eval $$my_field{$f,'url'};
        $value =~ s/^\s*//;
        $value =~ s/\s*$//;
        $length = length($value);
	if ( $@ ) {
	    &CgiError("Error in $f 's 'url' attribute :".
		      "<p>\n $@\nPlease report this error.\n");
            &dbi_disconnect($db_query, \$main_th);
	    goto BAIL;
	}
	$value = "<a href=\"$url\">$value</a>" if $url;
	$value .= " " x ($$my_field{$f,'length'} - $length);
	print "URL: $url" if $debug;
    }
    
    print "Value : $f : $val<br>\n " if $debug;
    return  $value;
}
#
#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#.PURPOSE	Evaluate an attribute using 'eval' - with error handling etc.
#
#.REMARKS	Dies on error.
#
#.RETURNS	The return value of the 'eval'
#------------------------------------------------------------------------------
sub eval_attr (\$\$\$)
{
    my($f, $attr, $val) = @_;
    my($return_code, $err_message);

    if ( defined $field{$f,$attr} ) {

	$return_code = eval $field{$f,$attr};
	if ( $@ ) {
	    $err_message =  "Error in $f"."'s  '$attr' attribute :<p>\n $@\n";
	    $err_message .=  "Please report this error.\n";
	    &CgiError($err_message);
	    goto BAIL;
	}
    } else {
	$return_code = $val;
    }
# line 3600 WDB
    return $return_code;
}
#
#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#.PURPOSE	Prints form header and title information
#------------------------------------------------------------------------------
sub write_header
{
    my ( $next_action, $header_text, $next_form, $header, $next_action );

    $header = $header_text = $next_action = $next_form = "";
    $next_form = $FORMNAME;

    if ( $ACTION eq "query_form" || $ACTION eq "form" ) {

        $header = ( $form{'Q_HEADER'} ) ? "$form{'Q_HEADER'}"
                            : "$form{'TITLE'} - Query Form";

        if ( $form{'Q_HTML'} ) {
            $header_text = "$form{'Q_HTML'}\n";
        } else {
            $header_text = "Please enter qualifiers in the fields ";
            $header_text .= "below and press the '<b>Search</b>' button.\n";
        }

        $next_action = "query";

        if ( $form{'RESULTS_FORM'} ) {
            $next_form = $form{'RESULTS_FORM'};
        }

    } elsif ( $ACTION eq "query" || $ACTION eq "page" ) {

        $header = ( $form{'R_HEADER'} ) ? "$form{'R_HEADER'}"
                                    : "$form{'TITLE'} - Query Result";
        $header_text = "$form{'R_HTML'}\n" if ( $form{'R_HTML'} );

    } elsif ( $ACTION eq "default" ) {

        $header = ( $form{'DEF_HEADER'} ) ? "$form{'DEF_HEADER'}"
			            : "$form{'TITLE'} - Default List";

        if ( $form{'DEF_HTML'} ) {
            $header_text = "$form{'DEF_HTML'}\n";
        } else {
            $header_text = 
	      "<H4>Press the 'Query' button to change qualifications.</h4>\n";
        }

# line 3650 WDB
    } elsif ( $ACTION eq "update_query_form" ) {

        $header = ( $form{'UQ_HEADER'} ) ? "$form{'UQ_HEADER'}"
				     : "$form{'TITLE'} - Update Query Form";

        if ( $form{'UQ_HTML'} ) {
            $header_text = "$form{'UQ_HTML'}\n";
        } else {
            $header_text = 
	      "Please enter qualifiers for the records you want to update".
	      " in the fields below and press the '<b>Search</b>' button.\n";
        }

        $next_action = "update_query";

        if ( $form{'RESULTS_FORM'} ) {
	    $next_form = $form{'RESULTS_FORM'};
	}
    
    } elsif ( $ACTION eq "insert_form" ) {

        $header = ( $form{'I_HEADER'} ) ? "$form{'I_HEADER'}" 
				    : "$form{'TITLE'} - Insert Form";

        if ( $form{'I_HTML'} ) {
  	    $header_text = "$form{'I_HTML'}\n";
        } else {
	    $header_text = "Please enter data in the fields".
                            " below and press the '<b>Insert</b>' button.\n".
	                    "<br>Fields marked with '<b>!</b>' are required.\n";
        }

        $next_action = "insert";

        if ( $form{'RESULTS_FORM'} ) {
	    $next_form = $form{'RESULTS_FORM'};
	}

    } elsif ( $ACTION eq "delete_query_form" ) {

        $header = ( $form{'DQ_HEADER'} ) ? "$form{'DQ_HEADER'}"
				     : "$form{'TITLE'} - Delete Query Form";
    
        if ( $form{'DQ_HTML'} ) {
            $header_text = "$form{'DQ_HTML'}\n";
        } else {
            $header_text = "Please enter qualifiers for the records you want ".
			   "to delete in the fields below and press the ".
			   "'<b>Search</b>' button.\n";
# line 3700 WDB
        }

        $next_action = "delete_query";

        if ( $form{'RESULTS_FORM'} ) {
	    $next_form = $form{'RESULTS_FORM'};
	}

    } elsif ( $ACTION eq "update_form" ) {

        $header = ( $form{'U_HEADER'} ) ? "$form{'U_HEADER'}" 
				    : "$form{'TITLE'} - Update Form";

        if ( $form{'U_HTML'} ) {
            $header_text = "$form{'U_HTML'}\n";
        } else {
            $header_text = "Please enter data in the fields below and press ".
			   "the '<b>Update</b>' button.\n".
                           "<br>Fields marked with '<b>!</b>' are required.\n";
        }

        $next_action = "update";

        if ( $form{'RESULTS_FORM'} ) {
	    $next_form = $form{'RESULTS_FORM'};
	}

    } elsif ( $ACTION eq "delete_form" ) {

        $header = ( $form{'D_HEADER'} ) ? "$form{'D_HEADER'}" 
				    : "$form{'TITLE'} - Delete Form";

        if ( $form{'D_HTML'} ) {
            $header_text = "$form{'D_HTML'}\n";
        } else {
            $header_text = "<BR>Please make sure this is the record you want ".
			   "to delete and press the '<b>Delete</b>' button.".
			   "<BR><BR>\n";
        }

        $next_action = "delete";

        if ( $form{'D_RESULTS_FORM'} ) {
            $next_form = $form{'D_RESULTS_FORM'};
        } elsif ( $form{'RESULTS_FORM'} ) {
	    $next_form = $form{'RESULTS_FORM'};
	}

    } elsif ( $ACTION eq "update_query" || $ACTION eq "delete_query" ) {
# line 3750 WDB

        $header = ( $form{'R_HEADER'} ) ? "$form{'R_HEADER'}" 
				    : "$form{'TITLE'} - Query Result";

        if ( $ACTION eq " update_query" ) {
	   $next_action = "update_form";
	} else {
	   $next_action = "delete_form";
	}

    } elsif ( $ACTION eq "insert" ) {

        $header = ( $form{'IR_HEADER'} ) ? "$form{'IR_HEADER'}"
				     : "$form{'TITLE'} - Insert Result";

        $header_text = "$form{'IR_HTML'}\n" if ( $form{'IR_HTML'} );

    } elsif ( $ACTION eq "update" ) {

        $header = ( $form{'UR_HEADER'} ) ? "$form{'UR_HEADER'}"
				     : "$form{'TITLE'} - Update Result";

        $header_text = "$form{'UR_HTML'}\n" if ( $form{'UR_HTML'} );

    } elsif ( $ACTION eq "delete" ) {

        if ( $form{'DR_HEADER'} ) {
            $header = "$form{'DR_HEADER'}";
        } else {
            $header = "$form{'TITLE'} - Delete Result";
        }

        $header_text = "$form{'DR_HTML'}\n" if ( $form{'DR_HTML'} );

    }

    print "<HTML><HEAD><TITLE>$form{'TITLE'}</TITLE></HEAD><BODY ";
    print "$form{'BODY_FORMAT'}" if defined $form{'BODY_FORMAT'};
    print ">\n";

    if ( $form{'DOCURL'} ) {
        $header = "<a href=\"$form{'DOCURL'}\">" . $header . "</a>";
    }

    print "<H1>$header</h1>\n" if $header;
    print "$header_text" if $header_text;

    print "<FORM ACTION=\"$WDB/$DBDIR/$next_form/$next_action\" ".
	  "METHOD=\"POST\">" if $next_action;
# line 3800 WDB

    if ( $ACTION eq "delete_form" && defined $form{'PASSWORD_DELETE'}) {

        print "<HR>The delete operation requires a username and password.<BR>";
        print "<BR><TABLE><TR>";
        print "<TH>Username: </TH>";
        print "<TD><INPUT NAME=\"username\" TYPE=text SIZE=8></TD>";
        print "<TD WIDTH=30>";
        print "<TH>Password: </TH>";
        print "<TD><INPUT NAME=\"password\" TYPE=password SIZE=8></TD>";
        print "</TR></TABLE>";
    }

    return 1;
}
#
#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#.PURPOSE	Generate the "menu" for the form, and result pages.
#
#.RETURNS	nothing
#------------------------------------------------------------------------------

sub write_menu
{
   my($add_menu, $form_buttons, $help_button, $query_button, $home_button);
   $add_menu = $form_buttons = $help_button = $query_button = $home_button = "";

    print "<hr>";

    if ( $ACTION eq "query_form" || $ACTION eq "update_query_form" ||
	 $ACTION eq "delete_query_form" || $ACTION eq "form" ) {

        $form_buttons = "<INPUT TYPE=\"submit\" VALUE=\"Search\">".
			" <INPUT TYPE=\"reset\" VALUE=\"Reset\">";

	$help_button = (defined $form{'HELP_BUTTON'}) ? 
	   " <A HREF=\"$form{'QUERY_HELP'}\"><img ALIGN=ABSMIDDLE ".
	   "src=\"$form{'HELP_BUTTON'}\" alt=\"[Help]\" ></A>" :
	   " <A HREF=\"$QUERY_HELP\"><img ALIGN=ABSMIDDLE ".
	   "src=\"$HELP_BUTTON\" alt=\"[Help]\" ></A>";

    } elsif ( $ACTION eq "query" || $ACTION eq "page" ) {

	if ($NO_QUERY_BUTTON == 0) {
	   if ( $form{'QUERY_FORM'} ) {
	      $query_button = 
	         " <a href=\"$WDB/$DBDIR/$form{'QUERY_FORM'}/query_form\">";
	   } else {
	      $query_button = " <a href=\"$WDB/$DBDIR/$FORMNAME/query_form\">";
# line 3850 WDB
	   }
	   $query_button .= (defined $form{'QUERY_BUTTON'}) ? 
	      " <img ALIGN=ABSMIDDLE src=\"$form{'QUERY_BUTTON'}\" ".
	      "alt=\"[Query]\" ></A>" :
	      " <img ALIGN=ABSMIDDLE src=\"$QUERY_BUTTON\" ".
	      "alt=\"[Query]\" ></A>";
        }

    } elsif ( $ACTION eq "insert_form" ) {

        $form_buttons = "<INPUT TYPE=\"submit\" VALUE=\"Insert\">".
			" <INPUT TYPE=\"reset\" VALUE=\"Reset\">";
	$help_button = (defined $form{'HELP_BUTTON'}) ? 
	   " <A HREF=\"$form{'QUERY_HELP'}\"><img ALIGN=ABSMIDDLE ".
	   "src=\"$form{'HELP_BUTTON'}\" alt=\"[Help]\" ></A>" :
	   " <A HREF=\"$QUERY_HELP\"><img ALIGN=ABSMIDDLE ".
	   "src=\"$HELP_BUTTON\" alt=\"[Help]\" ></A>";

    } elsif ( $ACTION eq "insert" ) {

	if ( $form{'QUERY_FORM'} ) {
	     $query_button =
		" <a href=\"$WDB/$DBDIR/$form{'QUERY_FORM'}/insert_form\">";
	} else {
	     $query_button = " <a href=\"$WDB/$DBDIR/$FORMNAME/insert_form\">";
	}
	$query_button .= (defined $form{'QUERY_BUTTON'}) ? 
	   " <img ALIGN=ABSMIDDLE src=\"$form{'QUERY_BUTTON'}\" ".
	   "alt=\"[Query]\" ></A>" :
	   " <img ALIGN=ABSMIDDLE src=\"$QUERY_BUTTON\" ".
	   "alt=\"[Query]\" ></A>";

    } elsif ( $ACTION eq "update_query" ) {

	if ( $form{'QUERY_FORM'} ) {
	     $query_button =
	      " <a href=\"$WDB/$DBDIR/$form{'QUERY_FORM'}/update_query_form\">";
	} else {
	     $query_button = 
	      " <a href=\"$WDB/$DBDIR/$FORMNAME/update_query_form\">";
	}
	$query_button .= (defined $form{'QUERY_BUTTON'}) ? 
	   " <img ALIGN=ABSMIDDLE src=\"$form{'QUERY_BUTTON'}\" ".
	   "alt=\"[Query]\" ></A>" :
	   " <img ALIGN=ABSMIDDLE ".
	   "src=\"$QUERY_BUTTON\" alt=\"[Query]\" ></A>";

    } elsif ( $ACTION eq "delete_query" ) {

# line 3900 WDB
	if ( $form{'QUERY_FORM'} ) {
	     $query_button = " <a href=\"$WDB/$DBDIR/$form{'QUERY_FORM'}/".
		   "delete_query_form\">";
	} else {
	     $query_button = 
		" <a href=\"$WDB/$DBDIR/$FORMNAME/delete_query_form\">";
	}
	$query_button .= (defined $form{'QUERY_BUTTON'}) ? 
	   " <img ALIGN=ABSMIDDLE src=\"$form{'QUERY_BUTTON'}\" ".
	   "alt=\"[Query]\" ></A>" :
	   " <img ALIGN=ABSMIDDLE ".
	   "src=\"$QUERY_BUTTON\" alt=\"[Query]\" ></A>";

    } elsif ( $ACTION eq "update_form" ) {

        $form_buttons = "<INPUT TYPE=\"submit\" VALUE=\"Update\">".
	   " <INPUT TYPE=\"reset\" VALUE=\"Reset\">";
	$help_button = (defined $form{'HELP_BUTTON'}) ? 
	   " <A HREF=\"$form{'QUERY_HELP'}\"><img ALIGN=ABSMIDDLE ".
	   "src=\"$form{'HELP_BUTTON'}\" alt=\"[Help]\" ></A>" :
	   " <A HREF=\"$QUERY_HELP\"><img ALIGN=ABSMIDDLE ".
	   "src=\"$HELP_BUTTON\" alt=\"[Help]\" ></A>";

    } elsif ( $ACTION eq "delete_form" ) {

        $form_buttons = "<INPUT TYPE=\"submit\" VALUE=\"Delete\">".
			" <INPUT TYPE=\"reset\" VALUE=\"Reset\">";
	$help_button = (defined $form{'HELP_BUTTON'}) ? 
	   " <A HREF=\"$form{'QUERY_HELP'}\"><img ALIGN=ABSMIDDLE ".
	   "src=\"$form{'HELP_BUTTON'}\" alt=\"[Help]\" ></A>" :
	   " <A HREF=\"$QUERY_HELP\"><img ALIGN=ABSMIDDLE ".
	   "src=\"$HELP_BUTTON\" alt=\"[Help]\" ></A>";

    } elsif ( $ACTION eq "update" ) {

	if ( $form{'RESULTS_FORM'} ) {
	     $query_button = " <a href=\"$WDB/$DBDIR/$form{'RESULTS_FORM'}/".
		   "update_query_form\">";
	} else {
	     $query_button = 
		" <a href=\"$WDB/$DBDIR/$FORMNAME/update_query_form\">";
	}
	$query_button .= (defined $form{'QUERY_BUTTON'}) ? 
	   " <img ALIGN=ABSMIDDLE src=\"$form{'QUERY_BUTTON'}\" ".
	   "alt=\"[Query]\" ></A>" :
	   " <img ALIGN=ABSMIDDLE ".
	   "src=\"$QUERY_BUTTON\" alt=\"[Query]\" ></A>";

    } elsif ( $ACTION eq "delete" ) {
# line 3950 WDB

        if ( $form{'D_RESULTS_FORM'} ) {
             $query_button = " <a href=\"$WDB/$DBDIR/$form{'D_RESULTS_FORM'}/".
		   "delete_query_form\">"
	} elsif ( $form{'RESULTS_FORM'} ) {
	     $query_button = " <a href=\"$WDB/$DBDIR/$form{'RESULTS_FORM'}/".
		   "delete_query_form\">";
	} else {
	     $query_button = 
		" <a href=\"$WDB/$DBDIR/$FORMNAME/delete_query_form\">";
	}
	$query_button .= (defined $form{'QUERY_BUTTON'}) ? 
	   " <img ALIGN=ABSMIDDLE src=\"$form{'QUERY_BUTTON'}\" ".
	   "alt=\"[Query]\" ></A>" :
	   " <img ALIGN=ABSMIDDLE ".
	   "src=\"$QUERY_BUTTON\" alt=\"[Query]\" ></A>";
    }

    print $form_buttons if $form_buttons;
    print $help_button if $help_button;
    print $query_button if $query_button;

    if (defined $form{'MAIN_MENU'}) {
	print " <a href=\"$form{'MAIN_MENU'}\">";
    } elsif ( $MAIN_MENU ) {
	print " <a href=\"$MAIN_MENU\">";
    }
    print "<img ALIGN=ABSMIDDLE alt=\"[Home]\" ";
    $home_button = (defined $form{'HOME_BUTTON'}) ? 
       "src=\"$form{'HOME_BUTTON'}\"></A>\n" :
       "src=\"$HOME_BUTTON\"></A>\n";
    print $home_button;
    print $user_menu if $user_menu;

    # Let the user add their own Perl functions at the top of the page.
    if ($form{'ADD_MENU'}) {
       $add_menu = eval $form{'ADD_MENU'};
       if ( $@ ) {
           &CgiError("Error in the form's ADD_MENU attribute :<p>\n $@\n");
           goto BAIL;
       }
       print $add_menu;
    }

    print "<hr>";
}

#------------------------------------------------------------------------------
# Wrotes a message at the end of a page.  Add your own here.
# line 4000 WDB
#------------------------------------------------------------------------------
sub write_tail
{
   &count_visitor;
}

#------------------------------------------------------------------------------
# Prints the number of records currently in the database and when the database
#  was last updated.  This information comes from a file that is updated when
#  the daily update script runs.
#------------------------------------------------------------------------------
sub count_records
{
    my($fname, $rec_count, $last_date, $return_code);

    $fname = "$form{'RECORD_COUNT_FILE'}";
    $return_code = open(RECS, "<$fname");
    if ($return_code == 0) {
       print "Could not open count file: $form{'RECORD_COUNT_FILE'}.<BR>";
       return(0);
    }
    while (<RECS>) {
        ($rec_count, $last_date) = split(/\|/, $_);
	chop $last_date;
        print "Records in database: <b>$rec_count</b>.";
        print " Last updated: <b>$last_date</b>.<BR>\n";
	last;
    }
    close(RECS);
}

#------------------------------------------------------------------------------
# Keeps a count of visitors to the WDB query page.
#------------------------------------------------------------------------------
sub count_visitor
{
    my($lock_file, $ip_address, $return_code, $current_count, $ip);
    $current_count = -1;

    return -1 if ! defined $form{'VISITOR_LOCK_FILE'};
    return -1 if ! defined $form{'VISITOR_COUNT_FILE'};
    $lock_file = "/tmp/$form{'VISITOR_LOCK_FILE'}";

    $ip_address = $ENV{'REMOTE_ADDR'};

    $return_code = &getlock($lock_file);
    if ($return_code == 0) { 
       print "Could not get lock file: $lock_file.<BR>\n"; 
       &release($lock_file);
# line 4050 WDB
       return -1;
    }

    if (!open(COUNT, "<$form{'VISITOR_COUNT_FILE'}")) {
       print "Could not open counter file: $form{'VISITOR_COUNT_FILE'}.<BR>";
       &release($lock_file);
       return -1;
    }

    while (<COUNT>) {
        $current_count = $_;
    }
    close(COUNT);

    # This ignores specified IP addresses when counting. Otherwise the count
    #  would skyrocket and not give an accurate count of visitors.
    if (defined @IP_IGNORE) {
       foreach $ip (@IP_IGNORE) {
          if ( $ip_address eq $ip ) {
	     &release($lock_file);
             print "<b>$form{'VISITOR_COUNT_TEXT'} $current_count</b><br>\n";
	     return 1;
          }
       }
    }

    if ($current_count > -1) {
       $current_count++;
       if (!open(COUNT, ">$form{'VISITOR_COUNT_FILE'}")) {
          print "Could not write to counter file: ".
	        "$form{'VISITOR_COUNT_FILE'}.<BR>";
          &release($lock_file);
          return -1;
       }
       print(COUNT $current_count);
       close(COUNT);
    }

    print "<b>$form{'VISITOR_COUNT_TEXT'} $current_count</b><br>\n";
    &release($lock_file);
    return 1;
}

#------------------------------------------------------------------------------
# Get an exclusive open on the lock file, so two visitors at the same time
#  don't clobber each other. Returns open file handle to tmp file. This code
#  is from Eric Murray's HTMLBBS.
#------------------------------------------------------------------------------
sub getlock (\$)
# line 4100 WDB
{
    my($lock_file) = @_;
    my($count);

    umask(0111);    # ensure html files will be world-writable.
                    # Ugly but required.

    for($count = 0; -f "$lock_file"; $count++) {
        # oh no. someone else is trying to update the message file.
        # so we wait.
        sleep(1);
        unlink "$lock_file" if ($count > 4);  # but not for long
    }

    open(TMP,">$lock_file") || return(0);
    return(1);
}

#------------------------------------------------------------------------------
# Releases lock. This code is from Eric Murray's HTMLBBS.
#------------------------------------------------------------------------------
sub release (\$)
{
    my($lock_file) = @_;
    close(TMP);
    unlink "$lock_file";  # unlock
    return(1);
}

#------------------------------------------------------------------------------
# Add a user menu to the regular one.  Use this to add button images, etc.
#------------------------------------------------------------------------------
sub add_menu (\$\$\$)
{
    my ($text, $href, $img) = @_;

    $user_menu .= " <a href=\"$href\">";
    if ( $img ) {
	$user_menu .= "<img alt=\"[$text]\" ALIGN=ABSMIDDLE src=\"$img\">";
    } else {
	$user_menu .= "[$text]";
    }
    $user_menu .= "</a>";
}

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#.PURPOSE       Convert SGML special Character codings into ASCII
#
#.RETURNS       The converted string
# line 4150 WDB
#------------------------------------------------------------------------------
sub SGMLtoASCII (\$)
{
    my ($str) = @_;
    $str =~ s/&gt;/>/g;
    $str =~ s/&lt;/=/g;
    return( $str );
}
#^L
#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#.PURPOSE       Encodes a string so it doesn't cause problems in URL.
#
#.RETURNS       The encoded string
#------------------------------------------------------------------------------
sub cgi_encode (\$)
{
    my ($str) = @_;
    print "Encoding string: $str<BR>\n" if $debug;
    $str = &escape($str,'[\x00-\x20"#%/+;<>?\x7F-\xFF]');
    $str =~ s/ /+/g;
    return( $str );
}
# ===========================================================================
# escape(): Return the passed string after replacing all characters matching
#           the passed pattern with their %XX hex escape chars.  Note that
#           the caller must be sure not to escape reserved URL characters
#           (e.g. / in pathnames, ':' between address and port, etc.) and thus
#           this routine can only be applied to each URL part separately. E.g.
#
#           $escname = &escape($name,'[\x00-\x20"#%/;<>?\x7F-\xFF]');
#------------------------------------------------------------------------------
sub escape (\$\$)
{
    my($str, $pat) = @_;

    $str =~ s/($pat)/sprintf("%%%02lx",unpack('C',$1))/ge;
    return($str);
}

# ===========================================================================
# unescape(): Return the passed URL after replacing all %NN escaped chars
#             with their actual character equivalents.
#------------------------------------------------------------------------------
sub unescape (\$)
{
    my($url) = @_;

    $url =~ s/%(..)/pack("C",hex($1))/ge;
    return $url;
# line 4200 WDB
}

#^L
#.COPYRIGHT NOTICE
# ==========================================================================
# Copyright Bo Frese Rasmussen 1994 - All Rights Reserved
# Copyright European Southern Observatory(ESO) 1994 - All Rights Reserved
# Copyright Space Telescope - European Coordinating Facility(ST-ECF) 1994
# - All Rights Reserved
#
# Permission to use, copy, modify, and distribute this software and its
# documentation for any purpose is hereby granted without fee, provided
# that the above copyright notice appear in all copies and that both that
# copyright notice, this permission notice, and the following disclaimer
# appear in supporting documentation, and that the names of the copyright
# holders, not be used in advertising or publicity pertaining to
# distribution of the software without specific, written prior
# permission.
#
# BO FRESE RASMUSSEN, EUROPEAN SOUTHERN OBSERVATORY, AND OTHER
# CONTRIBUTORS OF THIS SOFTWARE DISCLAIM ALL WARRANTIES WITH REGARD TO
# THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS FOR ANY PARTICULAR PURPOSE.  IN NO EVENT SHALL BO FRESE
# RASMUSSEN, EUROPEAN SOUTHERN OBSERVATORY OR ANY OTHER CONTRIBUTOR BE
# LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY
# DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA, OR PROFITS,
# WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE, OR OTHER TORTIOUS ACTION,
# ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
# SOFTWARE.
#------------------------------------------------------------------------------

# ---------------------------------------------------------------------------
# The escape and unescape functions are taken from the wwwurl.pl package
# developed by Roy Fielding <fielding@ics.uci.edu> as part of the Arcadia
# project at the University of California, Irvine. It is distributed
# under the Artistic License (included with your Perl distribution
# files).
# ---------------------------------------------------------------------------

# ---------------------------------------------------------------------------
# Additional code has been added to the original WDB by Jeff Rowe for Medix
#  Software Systems, Inc.  The new code is copyright (c) Medix Software 
#  Systems, Inc. 1997.  All rights reserved.  An unrestricted license is
#  granted to Jeff Rowe to distribute this code.  No warranty is expressed
#  or implied by Medix Software Systems or Jeff Rowe.
# ---------------------------------------------------------------------------
